import json
from queue import Queue
from threading import Event
from unittest import TestCase
from unittest.mock import ANY, MagicMock, patch

from websocket import WebSocketBadStatusException, WebSocketConnectionClosedException

from dakara_base.websocket_client import (
    AuthenticationError,
    NetworkError,
    NotConnectedError,
    ParameterError,
    WebSocketClient,
    connected,
)


class ConnectedTestCase(TestCase):
    """Test the `connected` decorator."""

    class Connected:
        def __init__(self):
            self.websocket = None

        @connected
        def dummy(self):
            pass

    def test_connected_sucessful(self):
        """Test the connected decorator when websocket is set.

        Use the `run` method for the test.
        """
        instance = self.Connected()

        # set the token
        instance.websocket = True

        # call a protected method
        instance.dummy()

    def test_connected_error(self):
        """Test the connected decorator when token is not set.

        Use the interal `get_token_header` method for test.
        """
        instance = self.Connected()

        # call a protected method
        with self.assertRaises(NotConnectedError):
            instance.dummy()


class WebSocketClientTestCase(TestCase):
    """Test the WebSocket connection with the server."""

    def setUp(self):
        # create a mock websocket
        self.websocket = MagicMock()

        # create a server URL
        self.url = "ws://www.example.com"

        # create a server WS endpoint URL
        self.url_ws = "ws://www.example.com/ws/"

        # create token header
        self.header = {"token": "token"}

        # create a reconnect interval
        self.reconnect_interval = 1

        # create stop event and errors queue
        self.stop = Event()
        self.errors = Queue()

        # create a DakaraServerWebSocketConnection instance
        self.client = WebSocketClient(
            self.stop,
            self.errors,
            {"url": self.url, "reconnect_interval": self.reconnect_interval},
            header=self.header,
            endpoint="ws/",
        )

    def set_websocket(self):
        """Set the websocket object to a mock."""
        self.client.websocket = MagicMock()

    @patch.object(WebSocketClient, "set_default_callbacks", autospec=True)
    def test_init_worker(self, mocked_set_default_callbacks):
        """Test the initialization."""
        # create the object
        client = WebSocketClient(self.stop, self.errors, {"url": self.url})

        # assert the call
        mocked_set_default_callbacks.assert_called_once_with(client)

        # use the already created client object
        self.assertEqual(self.client.server_url, self.url_ws)
        self.assertEqual(self.client.header, self.header)
        self.assertEqual(self.client.reconnect_interval, self.reconnect_interval)
        self.assertIsNone(self.client.websocket)

    def test_set_callback(self):
        """Test the assignation of a callback."""
        # create a callback function
        callback = MagicMock()

        # pre assert the callback is not set yet
        self.assertIsNot(self.client.callbacks.get("test"), callback)

        # call the method
        self.client.set_callback("test", callback)

        # post assert the callback is now set
        self.assertIs(self.client.callbacks.get("test"), callback)

    @patch.object(WebSocketClient, "abort", autospec=True)
    def test_exit_worker(self, mocked_abort):
        """Test to exit the worker."""
        # call the method
        with self.assertLogs("dakara_base.websocket_client", "DEBUG") as logger:
            self.client.exit_worker()

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            ["DEBUG:dakara_base.websocket_client:Aborting websocket connection"],
        )

        # assert the call
        mocked_abort.assert_called_with(self.client)

    @patch.object(WebSocketClient, "on_connected", autospec=True)
    def test_on_open(self, mocked_on_connected):
        """Test the callback on connection open."""
        # call the method
        with self.assertLogs("dakara_base.websocket_client", "DEBUG") as logger:
            self.client.on_open()

        # assert the effect
        self.assertFalse(self.client.retry)

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            ["INFO:dakara_base.websocket_client:Websocket connected to server"],
        )

        # assert the call
        mocked_on_connected.assert_called_with(self.client)

    @patch.object(WebSocketClient, "create_timer")
    @patch.object(WebSocketClient, "on_connection_lost")
    def test_on_close_normal(self, mocked_on_connection_lost, mocked_create_timer):
        """Test the callback on connection close when the program is closing."""
        # create the websocket
        self.set_websocket()

        # pre assert
        self.assertIsNotNone(self.client.websocket)
        self.assertFalse(self.client.retry)

        # set the program is closing
        self.stop.set()

        # call the method
        with self.assertLogs("dakara_base.websocket_client", "DEBUG") as logger:
            self.client.on_close(999, "reason")

        # assert the websocket object has been destroyed
        self.assertIsNone(self.client.websocket)
        self.assertFalse(self.client.retry)

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_base.websocket_client:Code 999: reason",
                "INFO:dakara_base.websocket_client:Websocket disconnected from server",
            ],
        )

        # assert the call
        mocked_create_timer.assert_not_called()
        mocked_on_connection_lost.assert_not_called()

    @patch.object(WebSocketClient, "create_timer", autospec=True)
    @patch.object(WebSocketClient, "on_connection_lost", autospec=True)
    def test_on_close_retry(self, mocked_on_connection_lost, mocked_create_timer):
        """Test the callback on connection close when connection should retry."""
        # create the websocket
        self.set_websocket()

        # pre assert
        self.assertFalse(self.client.retry)

        # call the method
        with self.assertLogs("dakara_base.websocket_client", "DEBUG") as logger:
            self.client.on_close(None, None)

        # assert the retry flag is set
        self.assertTrue(self.client.retry)

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            [
                "ERROR:dakara_base.websocket_client:Websocket connection lost",
                "WARNING:dakara_base.websocket_client:Trying to reconnect in 1 s",
            ],
        )

        # assert the different calls
        mocked_create_timer.assert_called_with(
            self.client, self.reconnect_interval, self.client.run
        )
        mocked_create_timer.return_value.start.assert_called_with()
        mocked_on_connection_lost.assert_called_with(self.client)

    @patch.object(WebSocketClient, "receive_dummy", create=True)
    def test_on_message_successful(self, mocked_receive_dummy):
        """Test a normal use of the on message method."""
        event = '{"type": "dummy", "data": "data"}'
        content = "data"

        # call the method
        self.client.on_message(event)

        # assert the dummy method has been called
        mocked_receive_dummy.assert_called_with(content)

    def test_on_message_failed_json(self):
        """Test the on message method when event is not a JSON string."""
        event = "definitely not a JSON string"

        # call the method
        with self.assertLogs("dakara_base.websocket_client", "DEBUG") as logger:
            self.client.on_message(event)

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            [
                "ERROR:dakara_base.websocket_client:"
                "Unexpected message from the server: '{}'".format(event)
            ],
        )

    def test_on_message_failed_type(self):
        """Test the on message method when event has an unknown type."""
        event = '{"type": "dummy", "data": "data"}'

        # call the method
        with self.assertLogs("dakara_base.websocket_client", "DEBUG") as logger:
            self.client.on_message(event)

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            [
                "ERROR:dakara_base.websocket_client:"
                "Event of unknown type received 'dummy'"
            ],
        )

    def test_on_message_failed_no_type(self):
        """Test the on message method when event has no type."""
        event = '{"data": "data"}'

        # call the method
        with self.assertLogs("dakara_base.websocket_client", "DEBUG") as logger:
            self.client.on_message(event)

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            ["ERROR:dakara_base.websocket_client:Event of no type received"],
        )

    def test_on_error_closing(self):
        """Test the callback on error when the program is closing.

        The error should be ignored.
        """
        # close the program
        self.stop.set()

        # call the method
        self.client.on_error(Exception("error message"))

        # assert the list of errors is empty
        self.assertTrue(self.errors.empty())

    def test_on_error_unknown(self):
        """Test the callback on an unknown error.

        The error should be logged only.
        """
        # pre assert
        self.assertFalse(self.stop.is_set())

        class CustomError(Exception):
            pass

        # call the method
        with self.assertLogs("dakara_base.websocket_client", "DEBUG") as logger:
            self.client.on_error(CustomError("error message"))

        # assert the list of errors is empty
        self.assertTrue(self.errors.empty())

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            ["ERROR:dakara_base.websocket_client:Websocket: error message"],
        )

    def test_on_error_authentication(self):
        """Test the callback on error when the authentication is refused."""
        # pre assert
        self.assertFalse(self.stop.is_set())

        # call the method
        self.client.on_error(WebSocketBadStatusException("error %s %s", 0))

        # assert the list of errors is not empty
        self.assertFalse(self.errors.empty())
        _, error, _ = self.errors.get()
        self.assertIsInstance(error, AuthenticationError)

    def test_on_error_network_normal(self):
        """Test the callback on error when the server is unreachable."""
        # pre assert
        self.assertFalse(self.client.retry)
        self.assertFalse(self.stop.is_set())

        # call the method
        self.client.on_error(ConnectionRefusedError("error"))

        # assert the list of errors is not empty
        self.assertFalse(self.errors.empty())
        _, error, _ = self.errors.get()
        self.assertIsInstance(error, NetworkError)

    def test_on_error_network_retry(self):
        """Test the callback on error when the server is unreachable on retry.

        No exception should be raised, the error should be logged only.
        """
        # pre assert
        self.assertFalse(self.stop.is_set())

        # set retry flag on
        self.client.retry = True

        # call the method
        with self.assertLogs("dakara_base.websocket_client", "DEBUG") as logger:
            self.client.on_error(ConnectionRefusedError("error"))

        # assert the list of errors is empty
        self.assertTrue(self.errors.empty())

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            ["WARNING:dakara_base.websocket_client:Unable to talk to the server"],
        )

    def test_on_error_endpoint(self):
        """Test the callback on error when the endpoint is invalid."""
        # pre assert
        self.assertFalse(self.stop.is_set())

        # call the method
        self.client.on_error(ConnectionResetError("error"))

        # assert the call
        self.assertFalse(self.errors.empty())
        _, error, _ = self.errors.get()
        self.assertIsInstance(error, ParameterError)

    def test_on_error_closed(self):
        """Test the callback on error when the connection is closed by server."""
        # pre assert
        self.assertFalse(self.stop.is_set())
        self.assertFalse(self.client.retry)

        # call the methods
        self.client.on_error(WebSocketConnectionClosedException("error"))

    def test_send_no_data(self):
        """Test to send message without data."""
        event = '{"type": "my_type"}'
        message_type = "my_type"

        # set the websocket
        self.set_websocket()

        # call the method
        self.client.send(message_type)

        # assert the call
        # we have to parse the strings, as the order in the dictionary is not
        # guaranteed in Python < 3.6
        self.client.websocket.send.assert_called_with(ANY)
        _, args, _ = self.client.websocket.send.mock_calls[0]
        event_sent = args[0]
        self.assertEqual(json.loads(event), json.loads(event_sent))

    def test_send_data_falsy(self):
        """Test to send message with falsy data."""
        event = '{"type": "my_type", "data": 0}'
        message_type = "my_type"

        # set the websocket
        self.set_websocket()

        # call the method
        self.client.send(message_type, 0)

        # assert the call
        # we have to parse the strings, as the order in the dictionary is not
        # guaranteed in Python < 3.6
        self.client.websocket.send.assert_called_with(ANY)
        _, args, _ = self.client.websocket.send.mock_calls[0]
        event_sent = args[0]
        self.assertEqual(json.loads(event), json.loads(event_sent))

    def test_send_data(self):
        """Test to send message with data."""
        event = '{"type": "my_type", "data": [1, 2, 3]}'
        message_type = "my_type"
        data = [1, 2, 3]

        # set the websocket
        self.set_websocket()

        # call the method
        self.client.send(message_type, data)

        # assert the call
        # we have to parse the strings, as the order in the dictionary is not
        # guaranteed in Python < 3.6
        self.client.websocket.send.assert_called_with(ANY)
        _, args, _ = self.client.websocket.send.mock_calls[0]
        event_sent = args[0]
        self.assertEqual(json.loads(event), json.loads(event_sent))

    def test_abort_connected(self):
        """Test to abort the connection."""
        # pre assert
        self.assertFalse(self.client.retry)

        # set the websocket
        self.set_websocket()

        # call the method
        self.client.abort()

        # assert the call
        self.client.websocket.sock.abort.assert_called_with()
        self.assertFalse(self.client.retry)

    def test_abort_disconnected(self):
        """Test to abort the connection when already disconnected."""
        # pre assert
        self.assertFalse(self.client.retry)
        self.assertIsNone(self.client.websocket)

        # call the method
        self.client.abort()

        # assert the call
        self.assertFalse(self.client.retry)

    def test_abort_retry(self):
        """Test to abort the connection when retry is set."""
        # set the retry flag on
        self.client.retry = True

        # call the method
        self.client.abort()

        # assert the call
        self.assertFalse(self.client.retry)

    @patch.object(WebSocketClient, "on_error", autospec=True)
    @patch.object(WebSocketClient, "on_message", autospec=True)
    @patch.object(WebSocketClient, "on_close", autospec=True)
    @patch.object(WebSocketClient, "on_open", autospec=True)
    @patch("dakara_base.websocket_client.WebSocketApp", autospec=True)
    def test_run(
        self,
        mocked_websocket_app_class,
        mocked_on_open,
        mocked_on_close,
        mocked_on_message,
        mocked_on_error,
    ):
        """Test to create and run the connection."""
        # pre assert
        self.assertIsNone(self.client.websocket)

        # call the method
        with self.assertLogs("dakara_base.websocket_client", "DEBUG") as logger:
            self.client.run()

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            ["DEBUG:dakara_base.websocket_client:Preparing websocket connection"],
        )

        # assert the call
        mocked_websocket_app_class.assert_called_with(
            self.url_ws,
            header=self.header,
            on_open=ANY,
            on_close=ANY,
            on_message=ANY,
            on_error=ANY,
        )
        self.client.websocket.run_forever.assert_called_with()

        # assert that the callback are correctly set
        # since the callback methods are adapted, we cannot check directy if
        # the given method reference is the same as the corresponding instance
        # method
        # so, we check that calling the given method calls the instance method
        websocket = MagicMock()
        _, kwargs = mocked_websocket_app_class.call_args

        kwargs["on_open"](websocket)
        self.client.on_open.assert_called_with(self.client)

        kwargs["on_close"](websocket, None, None)
        self.client.on_close.assert_called_with(self.client, None, None)

        kwargs["on_message"](websocket, "message")
        self.client.on_message.assert_called_with(self.client, "message")

        kwargs["on_error"](websocket, "error")
        self.client.on_error.assert_called_with(self.client, "error")

        # post assert
        # in real world, this test is impossible, since the websocket object
        # has been destroyed by `on_close`
        # we use the fact this callback is not called to check if the
        # object has been created as expected
        # maybe there is a better scenario to test this
        self.assertIsNotNone(self.client.websocket)



from unittest import TestCase
from unittest.mock import ANY

from dakara_base.exceptions import (
    DakaraError,
    DakaraHandledError,
    generate_exception_handler,
    handle_all_exceptions,
)


class GenerateExceptionHandleTestCase(TestCase):
    def test_handler(self):
        """Test to generate and use an exception handler"""

        class MyError(Exception):
            pass

        handler = generate_exception_handler(MyError, "handler message")

        with self.assertRaisesRegex(MyError, r"initial message\nhandler message") as cm:
            with handler():
                raise MyError("initial message")

        self.assertIsInstance(cm.exception, MyError)
        self.assertIsInstance(cm.exception, DakaraHandledError)


class HandleAllExceptionsTestCase(TestCase):
    def test_normal_exit(self):
        """Test a normal exit."""
        with handle_all_exceptions("url") as exit_value:
            pass

        self.assertEqual(exit_value.value, 0)

    def test_keyboard_interrupt(self):
        """Test a Ctrl+C exit."""
        with self.assertLogs("dakara_base.exceptions") as logger:
            with handle_all_exceptions("url") as exit_value:
                raise KeyboardInterrupt

        self.assertEqual(exit_value.value, 255)
        self.assertListEqual(
            logger.output, ["INFO:dakara_base.exceptions:Quit by user"]
        )

    def test_keyboard_interrupt_already_caught(self):
        """Test a Ctrl+C exit that is caught by the executed block."""
        with handle_all_exceptions("url") as exit_value:
            try:
                raise KeyboardInterrupt

            except KeyboardInterrupt:
                pass

        self.assertEqual(exit_value.value, 0)

    def test_known_error(self):
        """Test a known error exit."""
        with self.assertLogs("dakara_base.exceptions") as logger:
            with handle_all_exceptions("url") as exit_value:
                raise DakaraError("error")

        self.assertEqual(exit_value.value, 1)
        self.assertListEqual(logger.output, ["CRITICAL:dakara_base.exceptions:error"])

    def test_known_error_debug(self):
        """Test a known error exit in debug mode."""
        with self.assertRaisesRegex(DakaraError, "error"):
            with handle_all_exceptions("url", debug=True) as exit_value:
                raise DakaraError("error")

        self.assertEqual(exit_value.value, 1)

    def test_unknown_error(self):
        """Test an unknown error exit."""
        with self.assertLogs("dakara_base.exceptions") as logger:
            with handle_all_exceptions("url") as exit_value:
                raise Exception("error")

        self.assertEqual(exit_value.value, 2)
        self.assertListEqual(
            logger.output,
            [ANY, "CRITICAL:dakara_base.exceptions:Please fill a bug report at 'url'"],
        )

    def test_unknown_error_debug(self):
        """Test an unknown error exit in debug mode."""
        with self.assertRaisesRegex(Exception, "error"):
            with handle_all_exceptions("url", debug=True) as exit_value:
                raise Exception("error")

        self.assertEqual(exit_value.value, 2)



from unittest import TestCase

from path import Path

from dakara_base.directory import AppDirsPath


class AppDirsPathTestCase(TestCase):
    def test_properties(self):
        appdirs = AppDirsPath("appname", "authorname")

        self.assertIsInstance(appdirs.site_config_dir, Path)
        self.assertIsInstance(appdirs.site_data_dir, Path)
        self.assertIsInstance(appdirs.user_cache_dir, Path)
        self.assertIsInstance(appdirs.user_config_dir, Path)
        self.assertIsInstance(appdirs.user_data_dir, Path)
        self.assertIsInstance(appdirs.user_documents_dir, Path)
        self.assertIsInstance(appdirs.user_log_dir, Path)
        self.assertIsInstance(appdirs.user_runtime_dir, Path)
        self.assertIsInstance(appdirs.user_state_dir, Path)



import os
from unittest import TestCase
from unittest.mock import PropertyMock, patch

try:
    from importlib.resources import path

except ImportError:
    from importlib_resources import path

from environs import Env
from path import Path
from yaml.parser import ParserError

from dakara_base.config import (
    AutoEnv,
    Config,
    ConfigInvalidError,
    ConfigNotFoundError,
    ConfigParseError,
    create_config_file,
    create_logger,
    set_loglevel,
)


class AutoEnvTestCase(TestCase):
    """Test the `AutoEnv` class."""

    def test_auto(self):
        """Test to parse some valid types."""
        env = AutoEnv()

        with patch.object(Env, "int") as mocked_int:
            env.auto(int, "aaa")
            mocked_int.assert_called_with("aaa")

        with patch.object(Env, "bool") as mocked_bool:
            env.auto(bool, "aaa")
            mocked_bool.assert_called_with("aaa")

        with patch.object(Env, "float") as mocked_float:
            env.auto(float, "aaa")
            mocked_float.assert_called_with("aaa")

        with patch.object(Env, "list") as mocked_list:
            env.auto(list, "aaa")
            mocked_list.assert_called_with("aaa")

    def test_auto_invalid(self):
        """Test to parse an invalid types."""
        env = AutoEnv()

        with self.assertRaises(AttributeError):
            env.auto(type(None), "aaa")

    def test_get(self):
        """Test to get a value."""
        env = AutoEnv()

        with patch.dict(os.environ, {"PREFIX_AAA": "my_val"}, clear=True):
            with env.prefixed("PREFIX_"):
                self.assertEqual(env.auto(str, "AAA"), "my_val")


class ConfigTestCase(TestCase):
    """Test the `Config` class."""

    def test_return_env_var(self):
        """Test return var env when present."""

        config = Config("dakara")

        # Add value
        config["server"] = "url"

        # Value can be accessed like a regular dict
        self.assertEqual(config.get("server"), "url")
        self.assertEqual(config["server"], "url")

        # Add a environment variable with the same name
        with patch.dict(os.environ, {"DAKARA_SERVER": "url_from_env"}, clear=True):
            # return value from environment variable
            self.assertEqual(config.get("server"), "url_from_env")
            self.assertEqual(config["server"], "url_from_env")

    def test_create_from_dict(self):
        """Test the creation from existing dict."""

        # Create nested dict
        config_raw = {
            "server": {"url": "http://a.b", "token": "gdfgdg"},
            "player": {"config1": "conf", "value": "testvalue"},
        }

        config = Config("DAKARA", config_raw)

        # Check child dicts were also converted
        self.assertIsInstance(config["server"], Config)
        self.assertIsInstance(config["player"], Config)

        # Add a environment variable corresponding to the url param
        with patch.dict(os.environ, {"DAKARA_SERVER_URL": "url_from_env"}, clear=True):
            # return value from environment variable
            self.assertEqual(config.get("server").get("url"), "url_from_env")
            self.assertEqual(config["server"]["url"], "url_from_env")

    def test_cast(self):
        """Test to cast values when getting them."""
        config = Config("DAKARA")

        with patch.dict(
            os.environ,
            {
                "DAKARA_BOOL": "yes",
                "DAKARA_INT": "42",
                "DAKARA_FLOAT": "3.1416",
                "DAKARA_STR": "abcd",
                "DAKARA_LIST": "item1,item2",
            },
            clear=True,
        ):
            self.assertTrue(config.get("bool", False))
            self.assertEqual(config.get("int", 1), 42)
            self.assertAlmostEqual(config.get("float", 1.1), 3.1416)
            self.assertEqual(config.get("str"), "abcd")
            self.assertListEqual(config.get("list", []), ["item1", "item2"])

    def test_set_iterable_reset(self):
        """Test setting an iterable erases previous stored data."""
        config = Config("DAKARA", {"val": True, "spy": True})
        config.set_iterable({"val": False})
        self.assertFalse(config["val"])
        self.assertNotIn("spy", config)

    def test_set_debug(self):
        """Test to set debug mode."""
        config = Config("DAKARA", {"loglevel": "INFO"})
        self.assertNotEqual(config["loglevel"], "DEBUG")
        config.set_debug()
        self.assertEqual(config["loglevel"], "DEBUG")

    @patch.object(Config, "check_mandatory_key")
    def test_check_madatory_keys(self, mocked_check_mandatory_key):
        """Test to check a list of keys."""
        config = Config("DAKARA")
        config.check_mandatory_keys(["key"])

        mocked_check_mandatory_key.assert_called_once_with("key")

    def test_check_madatory_key_missing(self):
        """Test to check config without a required key."""
        config = Config("DAKARA")

        with self.assertRaisesRegex(
            ConfigInvalidError, "Invalid config file, missing 'not-present'"
        ):
            config.check_mandatory_key("not-present")

    def test_load_file_success(self):
        """Test to load a config file."""
        config = Config("DAKARA")

        # call the method
        with self.assertLogs("dakara_base.config", "DEBUG") as logger:
            with path("tests.resources", "config.yaml") as file:
                config.load_file(Path(file))

        # assert the result
        self.assertEqual(config["key"]["subkey"], "value")

        # assert the effect on logs
        self.assertListEqual(
            logger.output,
            ["INFO:dakara_base.config:Loading config file '{}'".format(Path(file))],
        )

    def test_load_file_fail_not_found(self):
        """Test to load a not found config file."""
        config = Config("DAKARA")

        # call the method
        with self.assertLogs("dakara_base.config", "DEBUG"):
            with self.assertRaisesRegex(ConfigNotFoundError, "No config file found"):
                config.load_file(Path("nowhere"))

    @patch("dakara_base.config.yaml.safe_load", autospec=True)
    def test_load_file_fail_parser_error(self, mocked_safe_load):
        """Test to load an invalid config file."""
        # mock the call to yaml
        mocked_safe_load.side_effect = ParserError("parser error")

        config = Config("DAKARA")

        # call the method
        with self.assertLogs("dakara_base.config", "DEBUG"):
            with path("tests.resources", "config.yaml") as file:
                with self.assertRaisesRegex(
                    ConfigParseError, "Unable to parse config file"
                ):
                    config.load_file(Path(file))

    def test_config_env(self):
        """Test to load config and get value from environment."""
        config = Config("DAKARA")

        with self.assertLogs("dakara_base.config", "DEBUG"):
            with path("tests.resources", "config.yaml") as file:
                config.load_file(Path(file))

        self.assertNotEqual(config.get("key").get("subkey"), "myvalue")

        with patch.dict(os.environ, {"DAKARA_KEY_SUBKEY": "myvalue"}, clear=True):
            self.assertEqual(config.get("key").get("subkey"), "myvalue")


@patch("dakara_base.config.LOG_FORMAT", "my format")
@patch("dakara_base.config.LOG_LEVEL", "my level")
class CreateLoggerTestCase(TestCase):
    """Test the `create_logger` function."""

    @patch("dakara_base.progress_bar.progressbar.streams.wrap_stderr")
    @patch("dakara_base.config.coloredlogs.install", autospec=True)
    def test_normal(self, mocked_install, mocked_wrap_stderr):
        """Test to call the method normally."""
        # call the method
        create_logger()

        # assert the call
        mocked_install.assert_called_with(fmt="my format", level="my level")
        mocked_wrap_stderr.assert_not_called()

    @patch("dakara_base.progress_bar.progressbar.streams.wrap_stderr")
    @patch("dakara_base.config.coloredlogs.install", autospec=True)
    def test_wrap(self, mocked_install, mocked_wrap_stderr):
        """Test to call the method and request to wrap stderr."""
        # call the method
        create_logger(wrap=True)

        # assert the call
        mocked_install.assert_called_with(fmt="my format", level="my level")
        mocked_wrap_stderr.assert_called_with()

    @patch("dakara_base.progress_bar.progressbar.streams.wrap_stderr")
    @patch("dakara_base.config.coloredlogs.install", autospec=True)
    def test_custom(self, mocked_install, mocked_wrap_stderr):
        """Test to call the method with custom format and level."""
        # call the method
        create_logger(
            custom_log_format="my custom format", custom_log_level="my custom level"
        )

        # assert the call
        mocked_install.assert_called_with(
            fmt="my custom format", level="my custom level"
        )
        mocked_wrap_stderr.assert_not_called()


class SetLoglevelTestCase(TestCase):
    """Test the `set_loglevel` function."""

    @patch("dakara_base.config.coloredlogs.set_level", autospec=True)
    def test_configure_logger(self, mocked_set_level):
        """Test to configure the logger."""
        # call the method
        set_loglevel({"loglevel": "DEBUG"})

        # assert the result
        mocked_set_level.assert_called_with("DEBUG")

    @patch("dakara_base.config.coloredlogs.set_level", autospec=True)
    def test_configure_logger_no_level(self, mocked_set_level):
        """Test to configure the logger with no log level."""
        # call the method
        set_loglevel({})

        # assert the result
        mocked_set_level.assert_called_with("INFO")


@patch.object(Path, "copyfile")
@patch.object(Path, "exists")
@patch.object(Path, "mkdir_p")
@patch(
    "dakara_base.directory.AppDirsPath.user_config_dir",
    new_callable=PropertyMock(return_value=Path("path") / "to" / "directory"),
)
@patch(
    "dakara_base.config.path",
)
class CreateConfigFileTestCase(TestCase):
    """Test the config file creator."""

    def test_create_empty(
        self,
        mocked_path,
        mocked_user_config_dir,
        mocked_mkdir_p,
        mocked_exists,
        mocked_copyfile,
    ):
        """Test create the config file in an empty directory."""
        # setup mocks
        mocked_exists.return_value = False

        # call the function
        with self.assertLogs("dakara_base.config") as logger:
            create_config_file("module.resources", "config.yaml")

        # assert the call
        mocked_path.assert_called_with("module.resources", "config.yaml")
        mocked_mkdir_p.assert_called_with()
        mocked_exists.assert_called_with()
        mocked_copyfile.assert_called_with(
            Path("path") / "to" / "directory" / "config.yaml"
        )

        # assert the logs
        self.assertListEqual(
            logger.output,
            [
                "INFO:dakara_base.config:Config created in '{}'".format(
                    Path("path") / "to" / "directory" / "config.yaml"
                )
            ],
        )

    @patch("dakara_base.config.input")
    def test_create_existing_no(
        self,
        mocked_input,
        mocked_path,
        mocked_user_config_dir,
        mocked_mkdir_p,
        mocked_exists,
        mocked_copyfile,
    ):
        """Test create the config file in a non empty directory."""
        # setup mocks
        mocked_exists.return_value = True
        mocked_input.return_value = "no"

        # call the function
        create_config_file("module.resources", "config.yaml")

        # assert the call
        mocked_copyfile.assert_not_called()
        mocked_input.assert_called_with(
            "{} already exists, overwrite? [y/N] ".format(
                Path("path") / "to" / "directory" / "config.yaml"
            )
        )

    @patch("dakara_base.config.input")
    def test_create_existing_invalid_input(
        self,
        mocked_input,
        mocked_path,
        mocked_user_config_dir,
        mocked_mkdir_p,
        mocked_exists,
        mocked_copyfile,
    ):
        """Test create the config file in a non empty directory with invalid input."""
        # setup mocks
        mocked_exists.return_value = True
        mocked_input.return_value = ""

        # call the function
        create_config_file("module.resources", "config.yaml")

        # assert the call
        mocked_copyfile.assert_not_called()

    @patch("dakara_base.config.input")
    def test_create_existing_force(
        self,
        mocked_input,
        mocked_path,
        mocked_user_config_dir,
        mocked_mkdir_p,
        mocked_exists,
        mocked_copyfile,
    ):
        """Test create the config file in a non empty directory with force overwrite."""
        # call the function
        create_config_file("module.resources", "config.yaml", force=True)

        # assert the call
        mocked_exists.assert_not_called()
        mocked_input.assert_not_called()
        mocked_copyfile.assert_called_with(
            Path("path") / "to" / "directory" / "config.yaml"
        )



from unittest import TestCase
from unittest.mock import ANY, MagicMock, patch

from requests.exceptions import RequestException

from dakara_base.http_client import (
    AuthenticationError,
    HTTPClient,
    MethodError,
    NotAuthenticatedError,
    ParameterError,
    ResponseInvalidError,
    ResponseRequestError,
    authenticated,
)


class AuthenticatedTestCase(TestCase):
    """Test the `authenticated` decorator."""

    class Authenticated:
        def __init__(self):
            self.token = None

        @authenticated
        def dummy(self):
            pass

    def test_authenticated_sucessful(self):
        """Test the authenticated decorator when token is set."""
        instance = self.Authenticated()

        # set the token
        instance.token = True

        # call a protected method
        instance.dummy()

    def test_authenticated_error(self):
        """Test the authenticated decorator when token is not set."""
        instance = self.Authenticated()

        # call a protected method
        with self.assertRaises(NotAuthenticatedError):
            instance.dummy()


class HTTPClientTestCase(TestCase):
    """Test the HTTP connection with a server."""

    def setUp(self):
        # create a token
        self.token = "token value"

        # create a server URL
        self.url = "http://www.example.com"

        # create a server API URL
        self.url_api = "http://www.example.com/api/"

        # create a server URL endpoint
        self.url_endpoint = "http://www.example.com/api/endpoint/"

        # create login URL
        self.url_login = "http://www.example.com/api/accounts/login/"

        # create a login and password
        self.login = "test"
        self.password = "test"

        # create a ServerHTTPConnection instance
        self.client = HTTPClient(
            {"url": self.url, "login": self.login, "password": self.password},
            endpoint_prefix="api/",
        )

    def set_token(self):
        """Set the token to the test client."""
        self.client.token = self.token

    def set_mute(self):
        """Set the client to mute communication errors when sending requests."""
        self.client.mute_raise = True

    def test_init(self):
        """Test to create object."""
        # use the already created client object
        self.assertEqual(self.client.server_url, self.url_api)
        self.assertEqual(self.client.login, self.login)
        self.assertEqual(self.client.password, self.password)
        self.assertIsNone(self.client.token)

    def test_load_missing_key(self):
        """Test to create object with missing mandatory key."""
        # try to create a client from invalid config
        client = HTTPClient({"url": self.url}, endpoint_prefix="api/")

        with self.assertRaisesRegex(
            ParameterError,
            "You have to either specify 'token' or the couple 'login' "
            "and 'password' in config file",
        ):
            client.load()

    @patch("dakara_base.http_client.requests.post", autospec=True)
    def test_send_request_raw_successful(self, mocked_post):
        """Test to send a raw request with the generic method."""
        # call the method
        with self.assertLogs("dakara_base.http_client", "DEBUG") as logger:
            self.client.send_request_raw(
                "post",
                "endpoint/",
                data={"content": "test"},
                message_on_error="error message",
            )

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_base.http_client:"
                "Sending POST request to http://www.example.com/api/endpoint/"
            ],
        )

        # assert the call
        mocked_post.assert_called_with(
            "http://www.example.com/api/endpoint/", data={"content": "test"}
        )

    def test_send_request_raw_error_method(self):
        """Test that a wrong method name fails for a generic raw request."""
        # call the method
        with self.assertRaises(MethodError):
            self.client.send_request_raw(
                "invalid",
                "endpoint/",
                data={"content": "test"},
                message_on_error="error message",
            )

    @patch("dakara_base.http_client.requests.post", autospec=True)
    def test_send_request_raw_error_request(self, mocked_post):
        """Test to send a raw request when there is a communication error."""
        # mock the response of the server
        mocked_post.side_effect = RequestException("error")

        # call the method
        with self.assertLogs("dakara_base.http_client", "DEBUG") as logger:
            with self.assertRaisesRegex(
                ResponseRequestError, "Error when communicating with the server: error"
            ):
                self.client.send_request_raw(
                    "post",
                    "endpoint/",
                    data={"content": "test"},
                    message_on_error="error message",
                )

        # assert the effect on logger
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_base.http_client:Sending POST request to "
                "http://www.example.com/api/endpoint/",
                "ERROR:dakara_base.http_client:error message, communication error",
            ],
        )

    @patch("dakara_base.http_client.requests.post", autospec=True)
    def test_send_request_raw_error_response(self, mocked_post):
        """Test to send a raw request when the response is invalid."""
        # mock the response of the server
        mocked_post.return_value.ok = False

        # call the method
        with self.assertLogs("dakara_base.http_client", "DEBUG"):
            with self.assertRaises(ResponseInvalidError):
                self.client.send_request_raw("post", "endpoint/")

    @patch("dakara_base.http_client.requests.post", autospec=True)
    def test_send_request_raw_error_response_custom(self, mocked_post):
        """Test to send a raw request when the response is invalid and an
        error function given."""
        # mock the response of the server
        mocked_post.return_value.ok = False

        # create error exception and error function
        class MyError(Exception):
            pass

        def on_error(response):
            return MyError()

        # call the method
        with self.assertLogs("dakara_base.http_client", "DEBUG"):
            with self.assertRaises(MyError):
                self.client.send_request_raw(
                    "post", "endpoint/", function_on_error=on_error
                )

    @patch.object(HTTPClient, "send_request_raw", autospec=True)
    def test_send_request_successful(self, mocked_send_request_raw):
        """Test to send a successful request."""
        # set the token
        self.set_token()

        # call the method
        self.client.send_request("post", "endpoint/", json={"key": "value"})

        # assert the call
        mocked_send_request_raw.assert_called_with(
            self.client,
            "post",
            "endpoint/",
            headers={"Authorization": "Token token value"},
            json={"key": "value"},
        )

    @patch.object(HTTPClient, "send_request_raw", autospec=True)
    def test_send_request_error_raised(self, mocked_send_request_raw):
        """Test to send an unsuccessful request which is not muted."""
        # set the token
        self.set_token()

        # raise an error
        mocked_send_request_raw.side_effect = ResponseInvalidError("invalid")

        # call the method
        with self.assertRaises(ResponseInvalidError):
            self.client.send_request("post", "endpoint/", json={"key": "value"})

    @patch.object(HTTPClient, "send_request_raw", autospec=True)
    def test_send_request_error_muted(self, mocked_send_request_raw):
        """Test to send an unsuccessful request which is muted."""
        # set the token
        self.set_token()
        self.set_mute()

        # raise an error
        mocked_send_request_raw.side_effect = ResponseInvalidError("invalid")

        # call the method
        response = self.client.send_request("post", "endpoint/", json={"key": "value"})

        # assert the response is None
        self.assertIsNone(response)

    @patch("dakara_base.http_client.requests.delete", autospec=True)
    @patch("dakara_base.http_client.requests.patch", autospec=True)
    @patch("dakara_base.http_client.requests.put", autospec=True)
    @patch("dakara_base.http_client.requests.post", autospec=True)
    @patch("dakara_base.http_client.requests.get", autospec=True)
    def test_methods(
        self, mocked_get, mocked_post, mocked_put, mocked_patch, mocked_delete
    ):
        """Test the different HTTP methods."""
        # set the token
        self.set_token()

        # mock the response
        response = MagicMock()
        response.json.return_value = "data"
        mocked_get.return_value = response
        mocked_post.return_value = response
        mocked_put.return_value = response
        mocked_patch.return_value = response
        mocked_delete.return_value = response

        for method in ("get", "post", "put", "patch", "delete"):
            # call the method

            response_obtained = getattr(self.client, method)("endpoint/")

            # assert the result
            self.assertEqual(response_obtained, "data")

        # assert the calls
        mocked_get.assert_called_with(self.url_endpoint, headers=ANY)
        mocked_post.assert_called_with(self.url_endpoint, headers=ANY)
        mocked_put.assert_called_with(self.url_endpoint, headers=ANY)
        mocked_patch.assert_called_with(self.url_endpoint, headers=ANY)
        mocked_delete.assert_called_with(self.url_endpoint, headers=ANY)

    @patch("dakara_base.http_client.requests.post", autospec=True)
    def test_authenticate_successful(self, mocked_post):
        """Test a successful authentication with the server."""
        # mock the response of the server
        mocked_post.return_value.ok = True
        mocked_post.return_value.json.return_value = {"token": self.token}

        # pre assertions
        self.assertIsNone(self.client.token)

        # call the method
        with self.assertLogs("dakara_base.http_client", "DEBUG") as logger:
            self.client.authenticate()

        # call assertions
        mocked_post.assert_called_with(
            self.url_login, json={"login": self.login, "password": self.password}
        )

        # post assertions
        self.assertIsNotNone(self.client.token)
        self.assertEqual(self.client.token, self.token)

        # assert effect on logger
        self.assertListEqual(
            logger.output,
            [
                "DEBUG:dakara_base.http_client:Authenticate to the server",
                "DEBUG:dakara_base.http_client:"
                "Sending POST request to http://www.example.com/api/accounts/login/",
                "INFO:dakara_base.http_client:Login to server successful",
                "DEBUG:dakara_base.http_client:Token: {}".format(self.token),
            ],
        )

    @patch("dakara_base.http_client.requests.post", autospec=True)
    def test_authenticate_token_present(self, mocked_post):
        """Test to bypass authentication with already present token."""
        # create token
        self.client.token = self.token

        # call the method
        self.client.authenticate()

        # call assertions
        mocked_post.assert_not_called()

        # post assertions
        self.assertEqual(self.client.token, self.token)

    @patch("dakara_base.http_client.requests.post", autospec=True)
    def test_authenticate_error_network(self, mocked_post):
        """Test a network error when authenticating."""
        # mock the response of the server
        mocked_post.side_effect = RequestException()

        # call the method
        with self.assertRaises(ResponseRequestError):
            with self.assertLogs("dakara_base.http_client", "DEBUG"):
                self.client.authenticate()

    @patch("dakara_base.http_client.requests.post", autospec=True)
    def test_authenticate_error_authentication(self, mocked_post):
        """Test an authentication error when authenticating."""
        # mock the response of the server
        mocked_post.return_value.ok = False
        mocked_post.return_value.status_code = 400

        # call the method
        with self.assertRaises(AuthenticationError):
            with self.assertLogs("dakara_base.http_client", "DEBUG"):
                self.client.authenticate()

    @patch("dakara_base.http_client.requests.post", autospec=True)
    def test_authenticate_error_other(self, mocked_post):
        """Test a server error when authenticating."""
        # mock the response of the server
        mocked_post.return_value.ok = False
        mocked_post.return_value.status_code = 999
        mocked_post.return_value.test = "error"

        # call the method
        with self.assertRaises(AuthenticationError):
            with self.assertLogs("dakara_base.http_client", "DEBUG"):
                self.client.authenticate()

    def test_get_token_header(self):
        """Test the helper to get token header."""
        # set the token
        self.set_token()

        # call the method
        result = self.client.get_token_header()

        # call assertions
        self.assertEqual(result, {"Authorization": "Token " + self.token})

    def test_get_json_from_response(self):
        """Test the helper to get data from existing response."""
        # create the mock
        response = MagicMock()

        # call the method
        result = HTTPClient.get_json_from_response(response)

        # assert the result is not None
        self.assertIsNotNone(result)

        # assert the call
        response.json.assert_called_with()

    def test_get_json_from_response_none(self):
        """Test the helper to get data from no response."""
        # call the method
        result = HTTPClient.get_json_from_response(None)

        # assert the result is None
        self.assertIsNone(result)

    def test_get_json_from_response_no_content(self):
        """Test the helper to get data from a response with no content."""
        # create a response with no content
        response = MagicMock()
        response.text = ""
        response.json.side_effect = Exception("error")

        # call the method
        result = HTTPClient.get_json_from_response(response)

        # assert the result is None
        self.assertIsNone(result)



from unittest import TestCase
from unittest.mock import patch

from dakara_base.utils import URLParameterError, create_url, truncate_message


class TruncateMessageTestCase(TestCase):
    """Test the message display helper."""

    def test_small_message(self):
        """Test a small message is completelly displayed."""
        message = "few characters"
        message_displayed = truncate_message(message, limit=50)

        self.assertLessEqual(len(message_displayed), 50)
        self.assertEqual(message_displayed, "few characters")

    def test_long_message(self):
        """Test a long message is cut."""
        message = "few characters"
        message_displayed = truncate_message(message, limit=5)

        self.assertLessEqual(len(message_displayed), 5)
        self.assertEqual(message_displayed, "fe...")

    def test_too_short_limit(self):
        """Test a too short limit.

        This case is marginal.
        """
        message = "few characters"

        with self.assertRaises(AssertionError) as error:
            truncate_message(message, limit=2)

        self.assertEqual(str(error.exception), "Limit too short")


class CreateUrlTestCase(TestCase):
    """Test the URL creator helper."""

    def test_url(self):
        """Test to create URL directly with provided URL."""
        url = create_url(url="http://www.example.com", host="www.other.com")
        self.assertEqual(url, "http://www.example.com")

    def test_url_path(self):
        """Test to create URL directly with provided URL and path."""
        url = create_url(url="http://www.example.com", path="path/to/resource")
        self.assertEqual(url, "http://www.example.com/path/to/resource")

    def test_host(self):
        """Test to create URL with provided host."""
        url = create_url(host="www.example.com", scheme_no_ssl="http")
        self.assertEqual(url, "http://www.example.com")

    def test_host_ssl(self):
        """Test to create URL with provided host and SSL security."""
        url = create_url(host="www.example.com", ssl=True, scheme_ssl="https")
        self.assertEqual(url, "https://www.example.com")

    def test_host_port(self):
        """Test to create URL with provided host and port."""
        url = create_url(host="www.example.com", port=8000, scheme_no_ssl="http")
        self.assertEqual(url, "http://www.example.com:8000")

    def test_host_path(self):
        """Test to create URL with provided host and path."""
        url = create_url(
            host="www.example.com", path="path/to/resource", scheme_no_ssl="http"
        )
        self.assertEqual(url, "http://www.example.com/path/to/resource")

    def test_address_host(self):
        """Test to create URL with provided address containing host."""
        url = create_url(address="www.example.com", scheme_no_ssl="http")
        self.assertEqual(url, "http://www.example.com")

    def test_address_host_port(self):
        """Test to create URL with provided address containing host and port."""
        url = create_url(address="www.example.com:8000", scheme_no_ssl="http")
        self.assertEqual(url, "http://www.example.com:8000")

    def test_nothing(self):
        """Test to create URL when nothing is provided."""
        with self.assertRaises(URLParameterError):
            create_url()

    @patch("dakara_base.utils.furl", autospec=True)
    def test_invalid_furl(self, mocked_furl):
        """Test to create URL with invalid arguments for furl."""
        mocked_furl.side_effect = ValueError("error")
        with self.assertRaises(URLParameterError):
            create_url(host="www.example.com", scheme_no_ssl="http")



"""WebSocket client helper module.

This module provides the WebSocket client class WebSocketClient, using the
websocket_client library. The class is designed to work with servers sending
JSON messages. The message received from the server must be handled by custom
methods, consequently the class cannot be used directly:

>>> MyWebSocketClient(WebSocketClient):
...     def receive_new_song(self, data):
...         pass

The websocket client uses a token to authenticate to the server. The token can
be optained by the `HTTPClient` class from the `http_client` module, with the
`get_token_header` method. The client is a bit complex to setup:

>>> from theading import Event
>>> from queue import Queue
>>> stop = Event()
>>> errors = Queue()
>>> config = {
...     "url": "ws://www.example.com"
... }
>>> header = {
...     "Authorization": "Token my-token-value"
... }
>>> websocket_client = MyWebSocketClient(stop, errors, config,
...                                      endpoint="ws/", header=header)
>>> websocket_client.timer.start()
>>> websocket_client.timer.join()
"""


import json
import logging
from functools import wraps

from websocket import (
    WebSocketApp,
    WebSocketBadStatusException,
    WebSocketConnectionClosedException,
)

from dakara_base.exceptions import DakaraError
from dakara_base.safe_workers import WorkerSafeTimer, safe
from dakara_base.utils import create_url, truncate_message

logger = logging.getLogger(__name__)


RECONNECT_INTERVAL = 5


def connected(fun):
    """Decorator that ensures the websocket is set.

    It makes sure that the given function is called only if connected.
    If not connected, calling the function will raise a NotConnectedError.

    Args:
        fun (function): Function to decorate.

    Returns:
        function: Decorated function.
    """

    @wraps(fun)
    def call(self, *args, **kwargs):
        if self.websocket is None:
            raise NotConnectedError("No connection established")

        return fun(self, *args, **kwargs)

    return call


class WebSocketClient(WorkerSafeTimer):
    """WebSocket client.

    It communicates with WebSocket messages of a fixed format. The message is a
    JSON string representing a dictionary containing two keys: `type` (str)
    the type of the message, mandatory; and `data` (anything) the content of
    the message, optional. When sending a message, the data cannot be a lone
    None: in this case the key is not present in the message.

    On receiving a message, the client will call the method corresponding to
    its type, which name is `receive_<type of the message here>`. The method
    must accept one argument, which is the `data` key of the message.

    If the client is disconnected to the server, it tries to reconnect every
    `reconnect_interval` seconds.

    Being a `safe_workers.WorkerSafeTimer`, any non caught exception in
    callbacks will stop the entire program. Also, the class is a context
    manager which abort the connection on exit.

    Attributes:
        server_url (str): URL of the server.
        header (dict): Header to add to the HTTP requests for authentication.
        websocket (websocket.WebSocketApp): WebSocket connection object.
        retry (bool): Flag to retry a connection if it was lost.
        reconnect_interval (int): Interval in seconds between two reconnection
            attempts.
        callbacks (dict): Dictionary of extra callback actions to call on
            receiving messages.
        timer (threading.Timer): Timer used for reconnection.

    Args:
        config (dict): Configuration for the server, the same as
            DakaraServerHTTPConnection.
        endpoint (str): Enpoint of the WebSocket connection, added to the URL.
        header (dict): Header containing the authentication token.
    """

    def init_worker(self, config, endpoint="", header={}):
        # url
        self.server_url = create_url(
            **config, path=endpoint, scheme_no_ssl="ws", scheme_ssl="wss"
        )

        # other
        self.header = header
        self.websocket = None
        self.retry = False
        self.reconnect_interval = config.get("reconnect_interval", RECONNECT_INTERVAL)

        # create callbacks
        self.callbacks = {}
        self.set_default_callbacks()

        # create timer
        self.timer = self.create_timer(0, self.run)

    def set_default_callbacks(self):
        """Stub for creating callbacks.

        The method is automatically called at initialization.
        """

    def exit_worker(self, *args, **kwargs):
        """Method called on exiting the worker to abort the connection."""
        logger.debug("Aborting websocket connection")
        self.abort()

    def set_callback(self, name, callback):
        """Assign an arbitrary callback.

        Callback is added to the `callbacks` dictionary attribute.

        Args:
            name (str): Name of the callback in the `callbacks` attribute.
            callback (function): Function to assign.
        """
        self.callbacks[name] = callback

    @safe
    def on_open(self):
        """Callback when the connection is open."""
        logger.info("Websocket connected to server")
        self.retry = False
        self.on_connected()

    @safe
    def on_close(self, code, reason):
        """Callback when the connection is closed.

        If the disconnection is not due to the end of the program, consider the
        connection has been lost. In that case, a reconnection will be
        attempted within `reconnect_interval` seconds.

        Args:
            code (int): Error code (often None).
            reason (str): Reason of the closed connection (often None).
        """
        if code or reason:
            logger.debug("Code %i: %s", code, reason)

        # destroy websocket object
        self.websocket = None

        if self.stop.is_set():
            logger.info("Websocket disconnected from server")
            return

        if not self.retry:
            logger.error("Websocket connection lost")

        self.retry = True
        self.on_connection_lost()

        # attempt to reconnect
        logger.warning("Trying to reconnect in %i s", self.reconnect_interval)
        self.timer = self.create_timer(self.reconnect_interval, self.run)
        self.timer.start()

    @safe
    def on_message(self, message):
        """Callback when a message is received.

        It will call the method which name corresponds to the event type, if
        possible. Methods must have the name `receive_<name of the event type
        here>`, the type must be indicated in the `type` key of the message.
        The content of the message must be in the `data` key.

        Any error is logged.

        Args:
            message (str): A JSON text of the event.
        """
        # convert the message to an event object
        try:
            event = json.loads(message)

        # if the message is not in JSON format, assume this is an error
        except json.JSONDecodeError:
            logger.error(
                "Unexpected message from the server: '%s'", truncate_message(message)
            )
            return

        # get the type of the message
        try:
            message_type = event["type"]

        except KeyError:
            logger.error("Event of no type received")
            return

        # attempt to call the corresponding method
        method_name = "receive_{}".format(message_type)
        try:
            getattr(self, method_name)(event.get("data"))

        except AttributeError:
            logger.error("Event of unknown type received '%s'", message_type)

    @safe
    def on_error(self, error):
        """Callback when an error occurs.

        Args:
            error (BaseException): Class of the error.

        Raises:
            AuthenticationError: If the connection is denied.
            NetworkError: If the communication cannot be established.
            ParameterError: If the endpoint is invalid.
        """
        # do not analyze error on program exit, as it will mistake the
        # WebSocketConnectionClosedException raised by invoking `abort` for a
        # server connection closed error
        if self.stop.is_set():
            return

        # the connection was refused
        if isinstance(error, WebSocketBadStatusException):
            raise AuthenticationError(
                "Unable to connect to server with this user"
            ) from error

        # the server is unreachable
        if isinstance(error, ConnectionRefusedError):
            if self.retry:
                logger.warning("Unable to talk to the server")
                return

            raise NetworkError("Network error, unable to talk to the server") from error

        # the requested endpoint does not exist
        if isinstance(error, ConnectionResetError):
            raise ParameterError("Invalid endpoint to the server") from error

        # connection closed by the server (see beginning of the method)
        if isinstance(error, WebSocketConnectionClosedException):
            # this case is handled by the on_close method
            return

        # other unlisted reason
        logger.error("Websocket: %s", str(error))

    def on_connected(self):
        """Custom callback when the connection is established with the server.

        This method is a stub that can be overloaded.
        """

    def on_connection_lost(self):
        """Custom callback when the connection is lost with the server.

        This method is a stub that can be overloaded.
        """

    @connected
    def send(self, message_type, data=None, *args, **kwargs):
        """Send data to the server.

        Convert it to JSON string before sending.

        Args:
            message_type (str): Type of the message.
            data (any): Serializable data to send.
            Other arguments are passed to `websocket.WebSocketApp.send`.
        """
        # add type to the content
        content = {"type": message_type}

        # add data to the content if any
        if data is not None:
            content["data"] = data

        return self.websocket.send(json.dumps(content), *args, **kwargs)

    def abort(self):
        """Request to interrupt the connection.

        Can be called from anywhere. It will raise a
        `WebSocketConnectionClosedException` which will be passed to
        `on_error`.
        """
        self.retry = False

        # if the connection is lost, the `websocket` object may not have the
        # `abort` method
        try:
            self.websocket.sock.abort()

        except AttributeError:
            pass

    def run(self):
        """Event loop.

        Create the websocket connection and wait events from it. The method can
        be interrupted with the `abort` method.

        The WebSocketApp class is a genki: It will never complain of anything.
        Wether it is unable to create a connection or its connection is lost,
        the `run_forever` method ends without any exception or non-None return
        value. Exceptions are handled by the yandere `on_error` callback.
        """
        logger.debug("Preparing websocket connection")
        self.websocket = WebSocketApp(
            self.server_url,
            header=self.header,
            on_open=lambda ws: self.on_open(),
            on_close=lambda ws, code, reason: self.on_close(code, reason),
            on_message=lambda ws, message: self.on_message(message),
            on_error=lambda ws, error: self.on_error(error),
        )
        self.websocket.run_forever()


class NotConnectedError(DakaraError):
    """Error raised when connection is missing."""


class AuthenticationError(DakaraError):
    """Error raised when authentication fails."""


class ParameterError(DakaraError, ValueError):
    """Error raised when server parameters are unproperly set."""


class NetworkError(DakaraError):
    """Error raised when unable to communicate with the server."""



"""HTTP client helper module.

This module provides the HTTP client class `HTTPClient`, built on the requests
library. The class is designed to be used with an API which communicates with
JSON messages.  It is pretty straightforward to use:

>>> config = {
...     "url": "http://www.example.com",
...     "login": "login here",
...     "password": "password here",
... }
>>> client = HTTPClient(config, endpoint_prefix="api/")
>>> client.authenticate()
>>> data = client.get("library/songs/")
>>> client.post("library/songs", json={"title": "some title"})
"""


import logging
from functools import wraps

import requests
from furl import furl

from dakara_base.exceptions import DakaraError
from dakara_base.utils import create_url, truncate_message

logger = logging.getLogger(__name__)


def authenticated(fun):
    """Decorator that ensures the token is set.

    It makes sure that the given function is called only if authenticated. If
    not authenticated, calling the function will raise a `NotAuthenticatedError`.

    Args:
        fun (function): Function to decorate.

    Returns:
        function: Decorated function.
    """

    @wraps(fun)
    def call(self, *args, **kwargs):
        if self.token is None:
            raise NotAuthenticatedError("No connection established")

        return fun(self, *args, **kwargs)

    return call


class HTTPClient:
    """HTTP client designed to work with an API.

    The API must use JSON for message content.

    The client uses a token credential policy only and authenticates with a
    traditional login/password mechanism. If a token is provided, it will be
    used without trying to authenticate.

    Attributes:
        AUTHENTICATE_ENDPOINT (str): Endpoint for authentication.
        mute_raise (bool): If true, no exception will be raised when performing
            connections with the server (but authentication), only logged.
        server_url (str): URL of the server.
        token (str): Value of the token. The token is set when successfuly
            calling `authenticate`.
        login (str): Login used for authentication.
        password (str): Password used for authentication.

    Args:
        config (dict): Config of the server.
        endpoint_prefix (str): Prefix of the endpoint, added to the URL.
        mute_raise (bool): If true, no exception will be raised when performing
            connections with the server (but authentication), only logged.

    Raises:
        ParameterError: If critical parameters cannot be found in the
            configuration.
    """

    AUTHENTICATE_ENDPOINT = "accounts/login/"

    def __init__(self, config, endpoint_prefix="", mute_raise=False):
        self.mute_raise = mute_raise

        # url
        self.server_url = create_url(**config, path=endpoint_prefix)

        # authentication
        self.token = config.get("token")
        self.login = config.get("login")
        self.password = config.get("password")

    def load(self):
        """Perform side effect actions.

        Raises:
            ParameterError: If there is neither a token or a couple
                login/password set.
        """
        if not self.token and not (self.login and self.password):
            raise ParameterError(
                "You have to either specify 'token' or the couple 'login' "
                "and 'password' in config file"
            )

    def send_request_raw(
        self,
        method,
        endpoint,
        *args,
        message_on_error="",
        function_on_error=None,
        **kwargs
    ):
        """Generic method to send requests to the server.

        It takes care of errors and raises exceptions.

        Args:
            method (str): Name of the HTTP method to use.
            endpoint (str): Endpoint to send the request to. Will be added to
                the end of the server URL.
            message_on_error (str): Message to display in logs in case of
                error. It should describe what the request was about.
            function_on_error (function): Fuction called if the request is not
                successful, it will receive the response and must return an
                exception that will be raised. If not provided, a basic error
                management is done.
            Extra arguments are passed to requests' get/post/put/patch/delete
                methods.

        Returns:
            requests.models.Response: Response object.

        Raises:
            MethodError: If the method is not supported.
            ResponseRequestError: For any error when communicating with the server.
            ResponseInvalidError: If the response has an error code different
                to 2**.
        """
        # handle method function
        if not hasattr(requests, method):
            raise MethodError("Method {} not supported".format(method))

        send_method = getattr(requests, method)

        # handle message on error
        if not message_on_error:
            message_on_error = "Unable to request the server"

        # forge URL
        url = furl(self.server_url).add(path=endpoint).url
        logger.debug("Sending %s request to %s", method.upper(), url)

        try:
            # send request to the server
            response = send_method(url, *args, **kwargs)

        except requests.exceptions.RequestException as error:
            # handle connection error
            logger.error("%s, communication error", message_on_error)
            raise ResponseRequestError(
                "Error when communicating with the server: {}".format(error)
            ) from error

        # return here if the request was made without error
        if response.ok:
            return response

        # otherwise call custom error management function
        if function_on_error:
            raise function_on_error(response)

        # otherwise manage error generically
        logger.error(message_on_error)
        logger.debug(
            "Error %i: %s", response.status_code, truncate_message(response.text)
        )

        raise ResponseInvalidError(
            "Error {} when communicationg with the server: {}".format(
                response.status_code, response.text
            )
        )

    @authenticated
    def send_request(self, *args, **kwargs):
        """Generic method to send requests to the server when connected.

        It adds token header for authentication and takes care of errors.
        If `mute_raise` is set, no exceptions are raised in case of error when
        communicating with the server.

        Args:
            See `send_request_raw`.

        Returns:
            requests.models.Response: Response object. None if an error occurs
            when communicating with the server and `mute_raise` is set.

        Raises:
            MethodError: If the method is not supported.
            ResponseRequestError: For any error when communicating with the server.
            ResponseInvalidError: If the response has an error code different
                to 2**.
        """
        try:
            # make the request
            return self.send_request_raw(
                *args, headers=self.get_token_header(), **kwargs
            )

        # manage request error
        except ResponseError:
            if self.mute_raise:
                return None

            raise

    def get(self, *args, **kwargs):
        """Generic method to get data on server.

        Args:
            endpoint (str): Endpoint to send the request to. Will be added to
                the end of the server URL.
            message_on_error (str): Message to display in logs in case of
                error. It should describe what the request was about.
            function_on_error (function): Fuction called if the request is not
                successful, it will receive the response and must return an
                exception that will be raised. If not provided, a basic error
                management is done.
            Extra arguments are passed to requests' get method.

        Returns:
            dict: Response object from the server.

        Raises:
            ResponseRequestError: For any error when communicating with the server.
            ResponseInvalidError: If the response has an error code different
                to 2**.
        """
        return self.get_json_from_response(self.send_request("get", *args, **kwargs))

    def post(self, *args, **kwargs):
        """Generic method to post data on server.

        Args:
            endpoint (str): Endpoint to send the request to. Will be added to
                the end of the server URL.
            message_on_error (str): Message to display in logs in case of
                error. It should describe what the request was about.
            function_on_error (function): Fuction called if the request is not
                successful, it will receive the response and must return an
                exception that will be raised. If not provided, a basic error
                management is done.
            Extra arguments are passed to requests' post method.


        Returns:
            dict: Response object from the server.

        Raises:
            ResponseRequestError: For any error when communicating with the server.
            ResponseInvalidError: If the response has an error code different
                to 2**.
        """
        return self.get_json_from_response(self.send_request("post", *args, **kwargs))

    def put(self, *args, **kwargs):
        """Generic method to put data on server.

        Args:
            endpoint (str): Endpoint to send the request to. Will be added to
                the end of the server URL.
            message_on_error (str): Message to display in logs in case of
                error. It should describe what the request was about.
            function_on_error (function): Fuction called if the request is not
                successful, it will receive the response and must return an
                exception that will be raised. If not provided, a basic error
                management is done.
            Extra arguments are passed to requests' put method.

        Returns:
            dict: Response object from the server.

        Raises:
            ResponseRequestError: For any error when communicating with the server.
            ResponseInvalidError: If the response has an error code different
                to 2**.
        """
        return self.get_json_from_response(self.send_request("put", *args, **kwargs))

    def patch(self, *args, **kwargs):
        """Generic method to patch data on server.

        Args:
            endpoint (str): Endpoint to send the request to. Will be added to
                the end of the server URL.
            message_on_error (str): Message to display in logs in case of
                error. It should describe what the request was about.
            function_on_error (function): Fuction called if the request is not
                successful, it will receive the response and must return an
                exception that will be raised. If not provided, a basic error
                management is done.
            Extra arguments are passed to requests' patch method.

        Returns:
            dict: Response object from the server.

        Raises:
            ResponseRequestError: For any error when communicating with the server.
            ResponseInvalidError: If the response has an error code different
                to 2**.
        """
        return self.get_json_from_response(self.send_request("patch", *args, **kwargs))

    def delete(self, *args, **kwargs):
        """Generic method to patch data on server.

        Args:
            endpoint (str): Endpoint to send the request to. Will be added to
                the end of the server URL.
            message_on_error (str): Message to display in logs in case of
                error. It should describe what the request was about.
            function_on_error (function): Fuction called if the request is not
                successful, it will receive the response and must return an
                exception that will be raised. If not provided, a basic error
                management is done.
            Extra arguments are passed to requests' delete method.

        Returns:
            dict: Response object from the server.

        Raises:
            ResponseRequestError: For any error when communicating with the server.
            ResponseInvalidError: If the response has an error code different
                to 2**.
        """
        return self.get_json_from_response(self.send_request("delete", *args, **kwargs))

    def authenticate(self):
        """Authenticate with the server.

        The authentication process relies on login/password which gives an
        authentication token. This token is stored in the instance.

        If a token was specified in the config, this function does nothing.

        Raises:
            See `send_request_raw`.
            AuthenticationError: If the connection is denied or if any onther
                error occurs.
        """

        if self.token:
            return

        data = {"login": self.login, "password": self.password}

        def on_error(response):
            # manage failed connection response
            if response.status_code == 400:
                return AuthenticationError(
                    "Login to server failed, check the config file"
                )

            # manage any other error
            return AuthenticationError(
                "Unable to authenticate to the server, error {code}: {message}".format(
                    code=response.status_code, message=truncate_message(response.text)
                )
            )

        # connect to the server with login/password
        logger.debug("Authenticate to the server")
        response = self.send_request_raw(
            "post",
            self.AUTHENTICATE_ENDPOINT,
            message_on_error="Unable to authenticate to the server",
            function_on_error=on_error,
            json=data,
        )

        # store token
        self.token = response.json().get("token")
        logger.info("Login to server successful")
        logger.debug("Token: %s", self.token)

    @authenticated
    def get_token_header(self):
        """Get the connection token as it should appear in the header.

        Can be called only after a successful authentication.

        Returns:
            dict: Formatted token.
        """
        return {"Authorization": "Token " + self.token}

    @staticmethod
    def get_json_from_response(response):
        """Parse the response of a request if possible.

        Args:
            response (requests.models.Response): Response of a request.

        Returns:
            dict: Parsed response. None if no response was given or response
            has no content.
        """
        if response and response.text:
            return response.json()

        return None


class ResponseError(DakaraError):
    """Generic error when communicating with the server."""


class ResponseRequestError(ResponseError):
    """Error when sending request to the server."""


class ResponseInvalidError(ResponseError):
    """Error with the request sent to the server."""


class AuthenticationError(ResponseInvalidError):
    """Error raised when authentication fails."""


class ParameterError(DakaraError, ValueError):
    """Error raised when server parameters are unproperly set."""


class MethodError(DakaraError, ValueError):
    """Error raised when using an unsupported method."""


class NotAuthenticatedError(DakaraError):
    """Error raised when authentication is missing."""



"""
Dakara base.

Collection of tools and helper modules for the Dakara Project.
"""

from dakara_base import (
    config,
    directory,
    exceptions,
    http_client,
    progress_bar,
    safe_workers,
    utils,
    version,
    websocket_client,
)
from dakara_base.version import __date__, __version__

__all__ = [
    "config",
    "directory",
    "exceptions",
    "http_client",
    "progress_bar",
    "safe_workers",
    "utils",
    "version",
    "websocket_client",
    "__version__",
    "__date__",
]



"""Directory helper module.

This module gives application name and project name:

>>> APP_NAME
... "dakara"
>>> PROJECT_NAME
... "DakaraProject"

It also gives an evolved version of `appdirs.AppDirs` that returns `path.Path`
objects:

>>> type(directories.user_config_dir)
... path.Path
"""
from path import Path
from platformdirs import PlatformDirs as AppDirs

APP_NAME = "dakara"
PROJECT_NAME = "DakaraProject"


class AppDirsPath(AppDirs):
    """AppDirs class that returns `path.Path` objects."""

    @property
    def site_config_dir(self):
        return Path(super().site_config_dir)

    @property
    def site_data_dir(self):
        return Path(super().site_data_dir)

    @property
    def user_cache_dir(self):
        return Path(super().user_cache_dir)

    @property
    def user_config_dir(self):
        return Path(super().user_config_dir)

    @property
    def user_data_dir(self):
        return Path(super().user_data_dir)

    @property
    def user_documents_dir(self):
        return Path(super().user_documents_dir)

    @property
    def user_log_dir(self):
        return Path(super().user_log_dir)

    @property
    def user_runtime_dir(self):
        return Path(super().user_runtime_dir)

    @property
    def user_state_dir(self):
        return Path(super().user_state_dir)


directories = AppDirsPath(APP_NAME, PROJECT_NAME, roaming=True)



import logging
import sys
from contextlib import contextmanager
from io import StringIO
from unittest import TestCase
from unittest.mock import MagicMock, patch

import progressbar

from dakara_base import progress_bar


class ShrinkablaTextWidgetTestCase(TestCase):
    """Test the widget for shrinkable text."""

    def test_no_shrink(self):
        """Test a not shrinked case."""
        # prepare mock objects
        progress = MagicMock()
        progress.term_width = 80
        data = MagicMock()

        # create the widget and update it
        widget = progress_bar.ShrinkableTextWidget("some text here")
        result = widget(progress, data)

        # assert the result
        self.assertEqual(len(result), 20)
        self.assertEqual(result, "some text here      ")

    def test_shrink(self):
        """Test a not shrinked case."""
        # prepare mock objects
        progress = MagicMock()
        progress.term_width = 40
        data = MagicMock()

        # create the widget and update it
        widget = progress_bar.ShrinkableTextWidget("some text here")
        result = widget(progress, data)

        # assert the result
        self.assertEqual(len(result), 10)
        self.assertEqual(result, "som...here")

    def test_too_short(self):
        """Test a too short case."""
        # create the widget
        with self.assertRaises(AssertionError) as error:
            progress_bar.ShrinkableTextWidget("some")

        # assert the error
        self.assertEqual(str(error.exception), "Text too short")


class ProgressBarTestCase(TestCase):
    """Test the default progress bar."""

    @staticmethod
    def get_lines(file):
        """Get a neat list of printed lines from a dummy file descriptor..

        Args:
            file (io.StringIO): Dummy file descriptor that get the output of
                the progress bar. It must not be closed.

        Returns:
            list of str: List of actually printed lines. Blank lines are removed.
        """
        content = file.getvalue()
        lines = content.replace("\r", "\n").splitlines()
        return [line for line in lines if line.strip()]

    def test_text(self):
        """Test a bar with text."""
        # call the bar
        with StringIO() as file:
            for _ in progress_bar.progress_bar(
                range(1), fd=file, term_width=65, text="some text here"
            ):
                pass

            lines = self.get_lines(file)

        # assert the lines
        self.assertEqual(len(lines), 3)
        self.assertEqual(
            lines,
            [
                "some text here     0% |    | Elapsed Time: 0:00:00 ETA:  --:--:--",
                "some text here   100% |####| Elapsed Time: 0:00:00 Time:  0:00:00",
                "some text here   100% |####| Elapsed Time: 0:00:00 Time:  0:00:00",
            ],
        )

    def test_no_text(self):
        """Test a bar without text."""
        # call the bar
        with StringIO() as file:
            for _ in progress_bar.progress_bar(range(1), fd=file, term_width=65):
                pass

            lines = self.get_lines(file)

        # assert the lines
        self.assertEqual(len(lines), 3)
        self.assertEqual(
            lines,
            [
                "  0% |                     | Elapsed Time: 0:00:00 ETA:  --:--:--",
                "100% |#####################| Elapsed Time: 0:00:00 Time:  0:00:00",
                "100% |#####################| Elapsed Time: 0:00:00 Time:  0:00:00",
            ],
        )

    def test_stderr_on_no_exception(self):
        """Test to check stderr is not captured if no exceptions occur."""
        stderr = StringIO()
        initial_stderr = sys.stderr

        # we patch `sys.stderr` as it is risky to work directly on it, and we
        # patch `progressbar.streams.original_stderr` as it is defined at
        # module loading
        with patch("sys.stderr", stderr), patch(
            "progressbar.streams.original_stderr", stderr
        ), wrap_stderr_progressbar():
            wrapped_stderr = sys.stderr

            # execute the progressbar without exception
            with StringIO() as file:
                for _ in progress_bar.progress_bar(range(1), fd=file, term_width=65):
                    pass

            sys.stderr.write("error")
            after_stderr = sys.stderr

        final_stderr = sys.stderr

        # assert stderrs
        self.assertIs(initial_stderr, final_stderr)
        self.assertIs(wrapped_stderr, after_stderr)
        self.assertIsNot(initial_stderr, wrapped_stderr)
        self.assertIsNot(stderr, wrapped_stderr)
        self.assertIsNot(stderr, initial_stderr)

        # assert stderr value
        value = stderr.getvalue()
        self.assertEqual(value, "error")

    def test_no_stderr_on_exception(self):
        """Test to check stderr does not remain captured after an exception.

        When leaving a progress bar by an exception, it does not call
        `progressbar.streams.stop_capturing` and the stderr is allways
        captured.
        """

        class MyException(Exception):
            pass

        stderr = StringIO()
        initial_stderr = sys.stderr

        with patch("sys.stderr", stderr), patch(
            "progressbar.streams.original_stderr", stderr
        ), wrap_stderr_progressbar():
            wrapped_stderr = sys.stderr

            # execute the progressbar with an exception
            with StringIO() as file:
                try:
                    for _ in progress_bar.progress_bar(
                        range(1), fd=file, term_width=65
                    ):
                        raise MyException("error")

                except MyException:
                    pass

            sys.stderr.write("error")
            after_stderr = sys.stderr

        final_stderr = sys.stderr

        # assert stderrs
        self.assertIs(initial_stderr, final_stderr)
        self.assertIs(wrapped_stderr, after_stderr)
        self.assertIsNot(initial_stderr, wrapped_stderr)
        self.assertIsNot(stderr, wrapped_stderr)
        self.assertIsNot(stderr, initial_stderr)

        # assert stderr value
        value = stderr.getvalue()
        self.assertEqual(value, "error")


@contextmanager
def wrap_stderr_progressbar():
    """Temporary wrap stderr with progressbar tools."""
    try:
        progressbar.streams.wrap_stderr()
        yield None

    finally:
        progressbar.streams.unwrap_stderr()


class NullBarTestCase(TestCase):
    """Test the null progress bar."""

    def setUp(self):
        # create logger similar to the one of the tested module
        self.logger = logging.getLogger("dakara_base.progress_bar")

    def test_text(self):
        """Test a null bar with text."""
        # call the bar
        with self.assertLogs("dakara_base.progress_bar") as logger:
            self.logger.info("start bar")
            for _ in progress_bar.null_bar(range(1), text="some text here"):
                pass

            self.logger.info("end bar")

        # assert the logs
        self.assertListEqual(
            logger.output,
            [
                "INFO:dakara_base.progress_bar:start bar",
                "INFO:dakara_base.progress_bar:some text here",
                "INFO:dakara_base.progress_bar:end bar",
            ],
        )

    def test_no_text(self):
        """Test a null bar without text."""
        # call the bar
        with self.assertLogs("dakara_base.progress_bar") as logger:
            self.logger.info("start bar")
            for _ in progress_bar.null_bar(range(1)):
                pass

            self.logger.info("end bar")

        # assert the logs
        self.assertListEqual(
            logger.output,
            [
                "INFO:dakara_base.progress_bar:start bar",
                "INFO:dakara_base.progress_bar:end bar",
            ],
        )



from contextlib import contextmanager
from queue import Queue
from threading import Event, Timer
from time import sleep
from unittest import TestCase
from unittest.mock import MagicMock

from dakara_base.safe_workers import (
    BaseSafeThread,
    BaseWorker,
    Runner,
    SafeThread,
    SafeTimer,
    UnredefinedThreadError,
    UnredefinedTimerError,
    Worker,
    WorkerSafeThread,
    WorkerSafeTimer,
    safe,
)


class MyError(Exception):
    """Dummy error class."""

    pass


class BaseTestCase(TestCase):
    """Generic test case.

    It includes some dummy functions a new assertion method.
    """

    def setUp(self):
        # create stop event and errors queue
        self.stop = Event()
        self.errors = Queue()

    def function_safe(self):
        """Function that does not raise any error."""
        return

    def function_error(self):
        """Function that raises a MyError."""
        raise MyError("test error")

    @contextmanager
    def assertNotRaises(self, ExceptionClass):
        """Assert that the provided exception does not raise.

        Args:
            ExceptionClass (class): Class of the exception.
        """
        try:
            yield None

        except ExceptionClass:
            self.fail("{} raised".format(ExceptionClass.__name__))


class SafeTestCase(BaseTestCase):
    """Test the `safe` decorator."""

    def create_classes(self):
        """Create dummy classes."""

        class Base:
            @safe
            def function_safe(self2):
                self.function_safe()

            @safe
            def function_error(self2):
                self.function_error()

        class Thread(BaseSafeThread, Base):
            pass

        class Worker(BaseWorker, Base):
            pass

        class Other(Base):
            pass

        return Thread, Worker, Other

    def test_worker_function_safe(self):
        """Test a safe function of a worker.

        Test that a non-error function does not trigger any error, does not set
        the stop event and does not put an error in the error queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create instance
        _, Worker, _ = self.create_classes()
        worker = Worker(self.stop, self.errors)

        # call the method
        worker.function_safe()

        # post assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

    def test_worker_function_error(self):
        """Test an error function of a worker.

        Test that an error function does not trigger any error, sets the stop
        event and puts a MyError in the error queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create instance
        _, Worker, _ = self.create_classes()
        worker = Worker(self.stop, self.errors)

        # call the method
        with self.assertNotRaises(MyError):
            worker.function_error()

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertFalse(self.errors.empty())
        _, error, _ = self.errors.get()
        self.assertIsInstance(error, MyError)

    def test_thread(self):
        """Test a thread.

        Test that a non-error function does not trigger any error, does not set
        the stop event and does not put an error in the error queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create instance
        Thread, _, _ = self.create_classes()
        thread = Thread(self.stop, self.errors)

        # call the method
        thread.function_safe()

        # post assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

    def test_other(self):
        """Test an other class.

        Test that the decorator raises an error, as the class is not supported.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create instance
        _, _, Other = self.create_classes()
        other = Other()

        # call the method
        with self.assertRaises(AssertionError):
            other.function_safe()

        # post assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())


class SafeThreadTestCase(BaseTestCase):
    """Test the SafeThread class."""

    def create_controlled_thread(self, target):
        """Helper to create a safe thread for a target function."""
        return SafeThread(self.stop, self.errors, target=target)

    def test_function_safe(self):
        """Test a safe function.

        Test that a non-error function run as a thread does not trigger any
        error, does not set the stop event and does not put an error in the
        error queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create thread
        controlled_thread = self.create_controlled_thread(self.function_safe)

        # run thread
        controlled_thread.start()
        controlled_thread.join()

        # post assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

    def test_function_error(self):
        """Test an error function.

        Test that an error function run as a thread does not trigger any error,
        sets the stop event and puts a MyError in the error queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create thread
        controlled_thread = self.create_controlled_thread(self.function_error)

        # run thread
        with self.assertNotRaises(MyError):
            controlled_thread.start()
            controlled_thread.join()

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertFalse(self.errors.empty())
        _, error, _ = self.errors.get()
        self.assertIsInstance(error, MyError)


class SafeTimerTestCase(SafeThreadTestCase):
    """Test the SafeTimer class."""

    def create_controlled_thread(self, target):
        """Helper to create a safe timer thread for a target function.

        The delay is non null (0.5 s).
        """
        return SafeTimer(self.stop, self.errors, 0.5, target)  # set a non-null delay


class WorkerTestCase(BaseTestCase):
    """Test the Worker class."""

    def test_run_safe(self):
        """Test a safe run.

        Test that a worker used with no error does not produce any error,
        finishes with a triggered stop event and an empty error queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with Worker(self.stop, self.errors):
            self.function_safe()

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertTrue(self.errors.empty())

    def test_run_error(self):
        """Test a run with error.

        Test that a worker used with error does produce an error, finishes
        with a triggered stop event and an empty error queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with self.assertRaises(MyError):
            with Worker(self.stop, self.errors):
                self.function_error()

        # there is no point continuing this test from here

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertTrue(self.errors.empty())

    def test_run_thread_safe(self):
        """Test a run with a safe thread.

        Test that a worker used with a non-error thread does not produce any
        error, finishes with a triggered stop event and an empty error queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with Worker(self.stop, self.errors) as worker:
            worker.thread = worker.create_thread(target=self.function_safe)
            worker.thread.start()
            worker.thread.join()

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertTrue(self.errors.empty())

    def test_run_thread_error(self):
        """Test a run with a thread with error.

        Test that a worker used with a thread with an error does not produce
        any error, finishes with a triggered stop event and a non-empty error
        queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with self.assertNotRaises(MyError):
            with Worker(self.stop, self.errors) as worker:
                worker.thread = worker.create_thread(target=self.function_error)
                worker.thread.start()
                worker.thread.join()

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertFalse(self.errors.empty())
        _, error, _ = self.errors.get()
        self.assertIsInstance(error, MyError)


class WorkerSafeTimerTestCase(BaseTestCase):
    """Test the WorkerSafeTimer class."""

    class WorkerSafeTimerToTest(WorkerSafeTimer):
        """Dummy worker class."""

        def function_already_dead(self):
            """Function that ends immediately."""
            return

        def function_to_cancel(self):
            """Function that calls itself in loop every second."""
            self.timer = Timer(1, self.function_to_cancel)
            self.timer.start()

        def function_to_join(self):
            """Function that waits one second."""
            sleep(1)

    def test_run_timer_dead(self):
        """Test to end a worker when its timer is dead.

        Test that a worker worker stopped with a dead timer finishes with a
        triggered stop event, an empty error queue and a still dead timer.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with self.WorkerSafeTimerToTest(self.stop, self.errors) as worker:
            worker.timer = worker.create_timer(0, worker.function_already_dead)
            worker.timer.start()
            worker.timer.join()

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertTrue(self.errors.empty())
        self.assertFalse(worker.timer.is_alive())

    def test_run_timer_cancelled(self):
        """Test to end a deamon when its timer is waiting.

        Test that a worker worker stopped with a waiting timer finishes with a
        triggered stop event, an empty error queue and a dead timer.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with self.WorkerSafeTimerToTest(self.stop, self.errors) as worker:
            worker.timer = worker.create_timer(0, worker.function_to_cancel)
            worker.timer.start()
            sleep(0.5)

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertTrue(self.errors.empty())
        self.assertFalse(worker.timer.is_alive())
        self.assertTrue(worker.timer.finished.is_set())

    def test_run_timer_joined(self):
        """Test to end a deamon when its timer is running.

        Test that a worker worker stopped with a running timer finishes with a
        triggered stop event, an empty error queue and a dead timer.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with self.WorkerSafeTimerToTest(self.stop, self.errors) as worker:
            worker.timer = worker.create_timer(0, worker.function_to_join)
            worker.timer.start()
            sleep(0.5)

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertTrue(self.errors.empty())
        self.assertFalse(worker.timer.is_alive())

    def test_unredifined_timer(self):
        """Test the timer must be redefined.

        Test that a worker worker with its default timer does not generate an
        error, but finishes with a triggered stop event and an non-empty error
        queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with self.assertNotRaises(UnredefinedTimerError):
            with self.WorkerSafeTimerToTest(self.stop, self.errors) as worker:
                worker.timer.start()

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertFalse(self.errors.empty())
        _, error, _ = self.errors.get()
        self.assertIsInstance(error, UnredefinedTimerError)


class WorkerSafeThreadTestCase(BaseTestCase):
    """Test the WorkerSafeThread class."""

    class WorkerSafeThreadToTest(WorkerSafeThread):
        """Dummy worker class."""

        def function_already_dead(self):
            """Function that ends immediately."""
            return

        def function_to_join(self):
            """Function that waits one second."""
            sleep(1)

    def test_run_thread_dead(self):
        """Test to end a worker when its thread is dead.

        Test that a worker worker stopped with a dead thread finishes with a
        triggered stop event, an empty error queue and a still dead thread.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with self.WorkerSafeThreadToTest(self.stop, self.errors) as worker:
            worker.thread = worker.create_thread(target=worker.function_already_dead)
            worker.thread.start()
            worker.thread.join()

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertTrue(self.errors.empty())
        self.assertFalse(worker.thread.is_alive())

    def test_run_thread_joined(self):
        """Test to end a deamon when its thread is running.

        Test that a worker worker stopped with a running thread finishes with a
        triggered stop event, an empty error queue and a dead thread.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with self.WorkerSafeThreadToTest(self.stop, self.errors) as worker:
            worker.thread = worker.create_thread(target=worker.function_to_join)
            worker.thread.start()
            sleep(0.5)

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertTrue(self.errors.empty())
        self.assertFalse(worker.thread.is_alive())

    def test_unredifined_thread(self):
        """Test the thread must be redefined.

        Test that a worker worker with its default thread does not generate an
        error, but finishes with a triggered stop event and an non-empty error
        queue.
        """
        # pre assertions
        self.assertFalse(self.stop.is_set())
        self.assertTrue(self.errors.empty())

        # create and run worker
        with self.assertNotRaises(UnredefinedThreadError):
            with self.WorkerSafeThreadToTest(self.stop, self.errors) as worker:
                worker.thread.start()

        # post assertions
        self.assertTrue(self.stop.is_set())
        self.assertFalse(self.errors.empty())
        _, error, _ = self.errors.get()
        self.assertIsInstance(error, UnredefinedThreadError)


class RunnerTestCase(BaseTestCase):
    """Test the Runner class.

    The class to test should leave because of a Ctrl+C, or because of an
    internal eror.
    """

    class WorkerNormal(Worker):
        """Dummy worker class."""

        def init_worker(self):
            self.thread = self.create_thread(target=self.test)

        def test(self):
            pass

    class WorkerError(Worker):
        """Dummy worker class that will fail."""

        def init_worker(self):
            self.thread = self.create_thread(target=self.test)

        def test(self):
            raise MyError("test error")

    def setUp(self):
        # create class to test
        self.runner = Runner()

    def test_run_safe_interrupt(self):
        """Test a run with an interruption by KeyboardInterrupt exception.

        The run should end with a set stop event and an empty errors queue.
        """
        # pre assertions
        self.assertFalse(self.runner.stop.is_set())
        self.assertTrue(self.runner.errors.empty())

        # modify stop event wait method
        self.runner.stop.wait = MagicMock()
        self.runner.stop.wait.side_effect = KeyboardInterrupt

        # call the method
        self.runner.run_safe(self.WorkerNormal)

        # post assertions
        self.assertTrue(self.runner.stop.is_set())
        self.assertTrue(self.runner.errors.empty())

        # assert stop event wait method was called
        self.runner.stop.wait.assert_called_once()

    def test_run_safe_error(self):
        """Test a run with an error.

        The run should raise a MyError, end with a set stop event and an
        empty error queue.
        """
        # pre assertions
        self.assertFalse(self.runner.stop.is_set())
        self.assertTrue(self.runner.errors.empty())

        # call the method
        with self.assertRaises(MyError):
            self.runner.run_safe(self.WorkerError)

        # post assertions
        self.assertTrue(self.runner.stop.is_set())
        self.assertTrue(self.runner.errors.empty())



"""Utils helper module.

This module regroups diverse helper functions.

The `truncate_message` function truncates a string if its width is langer than a
certain limit:

>>> string = "Lorem ipsum dolot sit amet."
>>> truncate_message(string, limit=15)
"Lorem ipsum..."

It was initialy designed to cut Django responses during development, as some
internal errors could make the server to respond by a very long HTML message,
polluting the logs.

The `create_url` is an URL creator build on top of the furl module. It is
typically designed to take a server config and forge an URL from it, wether the
URL is explicitally defined, or its components are individually defined, such
as host, port, etc.:

>>> config = {
...     "address": "www.example.com:8080",
...     "ssl": True,
...     "path": "api/",
... }
>>> create_url(**config)
"https://www.example.com:8080/api/"
"""
from furl import furl

from dakara_base.exceptions import DakaraError


def truncate_message(message, limit=100):
    """Display the first characters of a message.

    The message is truncated using ellipsis and stripped to avoid blank spaces
    before the ellipsis.

    Args:
        message (str): Message to truncate.
        limit (int): Maximum size of the message.

    Returns:
        str: Truncated message.
    """
    assert limit - 3 > 0, "Limit too short"

    if len(message) <= limit:
        return message

    return message[: limit - 3].strip() + "..."


def create_url(
    url="",
    address="",
    host="",
    port=None,
    path="",
    ssl=False,
    scheme_no_ssl="http",
    scheme_ssl="https",
    **kwargs
):
    """Create an URL from arguments.

    If `url` is given, the function returns it with `path` appended. If no
    `host` is given, `host` and `port` are extracted from `address` with the
    `host:port` format.  If `ssl` is given and True, `scheme_ssl` is used,
    otherewise `scheme_no_ssl` is used.  If `path` is given, it is appended to
    the URL.

    Args:
        url (str): Direct URL.
        address (str): Host, or host and port.
        host (str): Host.
        port (str): Port.
        path (str): Path appended to the URL.
        ssl (bool): Use a secured URL or not.
        scheme_no_ssl (str): Scheme used if `ssl` is false.
        scheme_ssl (str): Scheme used if `ssl` is true.
        Any other argument is ignored.

    Returns:
        str: URL string.

    Raises:
        URLParameterError: If `scheme` or `host` cannot be defined, or if the
            parameters are invalid to `furl.furl`.
    """
    # setting URL directly
    if url:
        return furl(url).add(path=path).url

    # getting host and port indirectly from address
    if not host:
        # try to separete host and port if they are both given in
        # address key in the form host:port
        try:
            host, port = address.split(":")

        except ValueError:
            host = address

    # getting scheme
    if ssl:
        scheme = scheme_ssl

    else:
        scheme = scheme_no_ssl

    # check mandatory arguments are given
    if not (scheme and host):
        raise URLParameterError(
            "Unable to set mandatory arguments for URL in server config, please check "
            "'url', 'address', 'host', 'port' and/or 'ssl'"
        )

    # combine the arguments
    try:
        return furl(scheme=scheme, host=host, port=port, path=path).url

    except ValueError as error:
        raise URLParameterError(
            "Error when setting URL in server config: {}".format(error)
        ) from error


class URLParameterError(DakaraError, ValueError):
    """Error raised when server parameters are unproperly set."""



"""Safe workers helper module.

This module provides some classes to facilitate the manipulaton of threads.
Especially, it guarantees that a failure in a thread will be notified in the
entire program.

Among them, the `SafeThread` and `SafeTimer` classes allow to retrieve
exceptions raised on the threaded/timed function. The `Worker` class can create
`SafeThread`s/`SafeTimer`s associated to itself. The `WorkerSafeThread` and
`WorkerSafeTimer` classes are context managers that ensure to close their
`SafeThread` or `SafeTimer` when exited. The `Runner` class can run the thread
of a `WorkerSafeThread` class untill internal error or user interrupt (Ctrl+C).

A main concept of this module is that a Python Event object is shared among the
instances of the classes of this module. This event is reffered as the stop
event, as it can be used to stop the program. A Python Queue object is also
shared among the different instances and allows to retreive exceptions raised
in sub-threads by the main one.

>>> from threading import Event
>>> from queue import Queue
>>> stop = Event()
>>> errors = Queue()
>>> worker = Worker(stop, errors)
>>> worker_with_thread = WorkerSafeThread(stop, errors)
>>> worker_with_timer = WorkerSafeTimer(stop, errors)
>>> worker.stop.set()
>>> worker_with_thread.stop.is_set()
True
"""


import logging
import sys
from functools import wraps
from queue import Empty, Queue
from threading import Event, Thread, Timer

from dakara_base.exceptions import DakaraError

logger = logging.getLogger(__name__)


def safe(fun):
    """Decorator to make the function safe.

    Any exception is caught and put in the error queue. This sets the stop
    event as well.

    The decorated function must be a method of a BaseSafeThread or a BaseWorker
    class (or inherited).
    """

    @wraps(fun)
    def call(self, *args, **kwargs):
        # check the target's class is a safe thread or a safe worker
        assert isinstance(self, (BaseSafeThread, BaseWorker)), (
            "The class '{}' of method '{}' is not a "
            "BaseSafeThread or a BaseWorker".format(
                self.__class__.__name__, fun.__name__
            )
        )

        # try to run the target
        try:
            return fun(self, *args, **kwargs)

        # if an error occurs, put it in the error queue and notify the stop
        # event
        except BaseException:
            self.errors.put_nowait(sys.exc_info())
            self.stop.set()

    return call


class BaseSafeThread:
    """Base class for thread executed within a Worker.

    The thread is connected to the stop event and the errors queue. In case of
    failure from the threaded function, the stop event is set and the exception
    is put in the error queue. The thread closes immediatlely.

    This mechanism allows to completely stop the execution of the program if an
    exception has been raised in a sub-thread. The excetpion is not shown on
    screen but passed to the main thread.

    This class is abstract and must be inherited with either `theading.Thread`
    or `threading.Timer`.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
    """

    def __init__(self, stop, errors, *args, **kwargs):
        # check arguments are valid
        assert isinstance(stop, Event), "Stop argument must be of type Event"
        assert isinstance(errors, Queue), "Errors argument must be of type Queue"

        # assign stop event and error queue
        self.stop = stop
        self.errors = errors

        # specific initialization
        super().__init__(*args, **kwargs)

    @safe
    def run(self):
        """Method to run as a thread safely."""
        return super().run()


class SafeThread(BaseSafeThread, Thread):
    """Thread executed within a Worker.

    The thread is connected to the stop event and the errors queue. In case of
    failure from the threaded function, the stop event is set and the exception
    is put in the error queue. The thread closes immediatlely.

    This mechanism allows to completely stop the execution of the program if an
    exception has been raised in a sub-thread. The excetpion is not shown on
    screen but passed to the main thread.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.

    Consult the help of `threading.Thread` for more information.
    """

    pass


class SafeTimer(BaseSafeThread, Timer):
    """Timer thread executed within a Workes.

    The timer thread is connected to the stop event and the errors queue. In
    case of failure from the threaded function, the stop event is set and the
    exception is put in the error queue. The timer thread closes immediatlely.

    This mechanism allows to completely stop the execution of the program if an
    exception has been raised in a timer sub-thread. The excetpion is not shown
    on screen but passed to the main thread.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.

    Consult the help of `threading.timer` for more information.
    """

    pass


class BaseWorker:
    """Base worker class.

    The base worker is bound to a stop event which when triggered will stop the
    program. It also has an errors queue to communicate errors to the main
    thread.

    It behaves like a context manager that returns itself on enter and triggers
    the stop event on exit.

    New threads should be created with the `create_thread` method and new
    thread timers with the `create_timer` method.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.

    Raises:
        AssertionError: If the `stop` or `errors` arguments are not
            respectivily Event and Queue.
    """

    def __init__(self, stop, errors):
        # associate the stop event
        assert isinstance(stop, Event), "Stop attribute must be of type Event"
        self.stop = stop

        # associate the errors queue
        assert isinstance(errors, Queue), "Errors attribute must be of type Queue"
        self.errors = errors

    def init_worker(self):
        """Custom init method stub."""
        pass

    def __enter__(self):
        """Simple context manager enter.

        Just call the custom enter method and returns the instance.
        """
        # call custom enter
        self.enter_worker()

        return self

    def enter_worker(self):
        """Custom enter method stub."""
        pass

    def __exit__(self, *args, **kwargs):
        """Simple context manager exit..

        Just triggers the stop event.
        """
        # notify the stop event
        self.stop.set()

    def exit_worker(self, *args, **kwargs):
        """Custom exit method stub."""
        pass

    def create_thread(self, *args, **kwargs):
        """Helper to easily create a SafeThread object..

        Args:
            See `threading.Thread`.

        Returns:
            SafeThread: Secured thread instance.
        """
        return SafeThread(self.stop, self.errors, *args, **kwargs)

    def create_timer(self, *args, **kwargs):
        """Helper to easily create a SafeTimer object..

        Args:
            See `threading.Timer`.

        Returns:
            SafeTimer: Secured timer thread instance.
        """
        return SafeTimer(self.stop, self.errors, *args, **kwargs)


class Worker(BaseWorker):
    """Worker class.

    The worker is bound to a stop event which when triggered will stop the
    program. It also has an errors queue to communicate errors to the main
    thread.

    It behaves like a context manager that returns itself on enter and triggers
    the stop event on exit.

    New threads should be created with the `create_thread` method and new
    thread timers with the `create_timer` method.

    Extra actions for context manager enter and exit should be put in the
    `enter_worker` and `exit_worker` methods.

    Initialization must be performed through the `init_worker` method. The
    initialization assigns the mandatory stop event and errors queue to the
    instance and calls the custom init method.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
    """

    def __init__(self, stop, errors, *args, **kwargs):
        super().__init__(stop, errors)

        # call custom init
        self.init_worker(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        """Simple context manager exit.

        Just triggers the stop event and call the custom exit method.
        """
        super().__exit__()

        logger.debug("Exiting worker method (%s)", self.__class__.__name__)

        # call custom exit
        self.exit_worker(*args, **kwargs)

        logger.debug("Exited worker method (%s)", self.__class__.__name__)


class WorkerSafeTimer(BaseWorker):
    """Worker class with safe timer.

    The worker class with safe timer is bound to a stop event which when
    triggered will stop the program. It also has an errors queue to communicate
    errors to the main thread.

    It contains a timer thread `timer` connected to a dummy function which must
    be redefined. New thread timers should be created with the `create_timer`
    method.

    It behaves like a context manager that gives itself on enter. On exit, it
    cancels and ends its timer thread and also triggers the stop event.

    Extra actions for context manager enter and exit should be put in the
    `enter_worker` and `exit_worker` methods.

    Initialization must be performed through the `init_worker` method.  The
    initialization assigns its own timer thread to the instance and makes it
    target a dummy method.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        timer (SafeTimer): Timer thread that must be redefined.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
    """

    def __init__(self, stop, errors, *args, **kwargs):
        super().__init__(stop, errors)

        # create timer for itself
        def redefine_me():
            """Dummy function that should not be used."""
            raise UnredefinedTimerError(
                "You must redefine the timer of a WorkerSafeTimer"
            )

        self.timer = self.create_timer(0, redefine_me)

        # perform other custom actions
        self.init_worker(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        """Worker context manager exit.

        Cancels and close the timer thread. It calls the custom context manager
        exit method.

        The stop event has already been triggered.
        """
        super().__exit__(*args, **kwargs)

        # exit now if the timer is not running
        if not self.timer.is_alive():
            return

        logger.debug(
            "Closing worker safe timer thread '%s' (%s)",
            self.timer.name,
            self.__class__.__name__,
        )

        # custom exit
        self.exit_worker(*args, **kwargs)

        # cancel the timer, if the timer was waiting
        self.timer.cancel()

        # wait for termination, if the timer was running
        self.timer.join()

        logger.debug(
            "Closed worker safe timer thread '%s' (%s)",
            self.timer.name,
            self.__class__.__name__,
        )


class WorkerSafeThread(BaseWorker):
    """Worker class with safe thread.

    The worker class with safe thread is bound to a stop event which when
    triggered will stop the program. It also has an errors queue to communicate
    errors to the main thread.

    It contains a thread `thread` connected to a dummy function which must de
    redefined. New threads should be created with the `create_thread` method.

    The instance is a context manager that gives itself on enter. On exit, it
    ends its own thread and also triggers the stop event.

    Extra actions for context manager enter and exit should be put in the
    `enter_worker` and `exit_worker` methods.

    Initialisation must be performed through the `init_worker` method. The
    initialization assigns its own thread to the instance and makes it target a
    dummy method.

    Attributes:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
        thread (SafeThread): Thread bound to the `run` method.

    Args:
        stop (threading.Event): Stop event that notify to stop the entire
            program when set.
        errors (queue.Queue): Error queue to communicate the exception to the
            main thread.
    """

    def __init__(self, stop, errors, *args, **kwargs):
        super().__init__(stop, errors)

        # create thread for itself
        def redefine_me():
            """Dummy function that should not be used."""
            raise UnredefinedThreadError(
                "You must redefine the thread of a WorkerSafeThread"
            )

        self.thread = self.create_thread(target=redefine_me)

        # perform other custom actions
        self.init_worker(*args, **kwargs)

    def __exit__(self, *args, **kwargs):
        """Worker context manager exit.

        Closes the thread. It calls the custom context manager exit method.

        The stop event has already been triggered.
        """
        super().__exit__(*args, **kwargs)

        # exit now if the thread is not running
        if not self.thread.is_alive():
            return

        logger.debug(
            "Closing worker safe thread '%s' (%s)",
            self.thread.name,
            self.__class__.__name__,
        )

        # custom exit action
        self.exit_worker(*args, **kwargs)

        # wait for termination
        self.thread.join()

        logger.debug(
            "Closed worker safe thread '%s' (%s)",
            self.thread.name,
            self.__class__.__name__,
        )


class Runner:
    """Runner class.

    The runner creates the stop event and errors queue. It is designed to
    execute the thread of a `WorkerSafeThread` instance until an error occurs
    or an user interruption pops out (Ctrl+C).

    The initialization creates the stop event and the errors queue and calls
    the custom init method.

    Attributes:
        POLLING_INTERVAL (float): For Windows only, interval between two
            attempts to wait for the stop event.
        stop (threading.Event): Stop event that notify to stop the execution of
            the thread.
        errors (queue.Queue): Error queue to communicate the exception of the
            thread.
    """

    POLLING_INTERVAL = 0.5

    def __init__(self, *args, **kwargs):
        # create stop event
        self.stop = Event()

        # create errors queue
        self.errors = Queue()

        # extra actions
        self.init_runner(*args, **kwargs)

    def init_runner(self, *args, **kwargs):
        """Custom initialization stub."""
        pass

    def run_safe(self, WorkerClass, *args, **kwargs):
        """Execute a WorkerSafeThread instance thread.

        The thread is executed and the method waits for the stop event to be
        set or a user interruption to be triggered (Ctrl+C).

        Args:
            WorkerClass (WorkerSafeThread): Worker class with safe thread.
                Note you have to pass a custom class based on
                `WorkerSafeThread`.
            Other arguments are passed to the thread of WorkerClass.
        """
        try:
            # create worker thread
            with WorkerClass(self.stop, self.errors, *args, **kwargs) as worker:

                logger.debug("Create worker thread")
                worker.thread.start()

                # wait for stop event
                logger.debug("Waiting for stop event")

                # We have to use a different code for Windows because the
                # Ctrl+C event will not be handled during `self.stop.wait()`.
                # This method is blocking for Windows, not for Linux, which is
                # due to the way Ctrl+C is differently handled by the two OSs.
                # For Windows, a quick and dirty solution consists in polling
                # the `self.stop.wait()` with a timeout argument, so the call
                # is non-permanently blocking.
                # More resources on this:
                # https://mail.python.org/pipermail/python-dev/2017-August/148800.html
                # https://stackoverflow.com/a/51954792/4584444
                if sys.platform.startswith("win"):
                    while not self.stop.is_set():
                        self.stop.wait(self.POLLING_INTERVAL)

                else:
                    self.stop.wait()

        # stop on Ctrl+C
        except KeyboardInterrupt:
            logger.debug("User stop caught")
            self.stop.set()

        # stop on error
        else:
            logger.debug("Internal error caught")

            # get the error from the error queue and re-raise it
            # a delay of 5 seconds is accorded for the error to be retrieved
            try:
                _, error, traceback = self.errors.get(5)
                error.with_traceback(traceback)
                raise error

            # if there is no error in the error queue, raise a general error
            # this case is very unlikely to happen and is not tested
            except Empty as empty_error:
                raise NoErrorCaughtError("Unknown error happened") from empty_error


class UnredefinedTimerError(DakaraError):
    """Unredefined timer error.

    Error raised if the default timer of the `WorkerSafeTimer` class has not
    been redefined.
    """


class UnredefinedThreadError(DakaraError):
    """Unredefined thread error.

    Error raised if the default thread of the `WorkerSafeTimer` class has not
    been redefined.
    """


class NoErrorCaughtError(RuntimeError):
    """No error caught error.

    Error raised if the safe workers mechanism stops for an error, but there is
    no error. This error is completely unexpected and hence does not inherit
    from DakaraError.
    """



"""Progress bar module.

This module regroups utilities for managing a progress bar.

If you use progress bar and logging at the same time, you should call
`dakara_base.config.create_logger` with `wrap=True`.

Two different bars are provided in the module.

The `progress_bar` is a progress bar that can display a descriptive text to
inform which task is going on:

>>> items = ["one", "two", "three"]
>>> for item in progress_bar(items, text="brief description of the task"):
...     pass

The text adapts to the screen width using the `ShrinkableTextWidget` widget.

The `null_bar` is a pseudo progress bar that does not display progress. It was
initialy designed to be used on logs with the same interface as a bar, in order
to not pollute the log file.

>>> items = ["one", "two", "three"]
>>> for item in null_bar(items, text="brief description of the task"):
...     pass

The text is displayed as a log entry.
"""


import logging

import progressbar
from progressbar.widgets import WidgetBase

logger = logging.getLogger(__name__)


class ShrinkableTextWidget(WidgetBase):
    """Widget which size auto-shrinks with terminal width.

    It contains a descriptive text using by default one quarter of the screen
    width, which can be truncated by the middle if it does not fit.

    Args:
        text (str): Text to display on screen.
        ratio (float): Ratio of screen width to use for text.
    """

    def __init__(self, text, ratio=0.25):
        super().__init__()

        assert len(text) > 5, "Text too short"
        self.text = text
        self.ratio = ratio

    def __call__(self, progress, data):
        # set widget width to a fraction of terminal width
        width = int(progress.term_width * self.ratio)

        # truncate text if necessary
        text = self.text
        if len(text) > width:
            half = int(width * 0.5)
            text = text[: half - 2].strip() + "..." + text[-half + 1 :].strip()

        return text.ljust(width)


def progress_bar(iterator, *args, text=None, **kwargs):
    """Generator that gives the default un-muted progress bar for the project.

    It prints an optionnal shrinkable text (if a text is provided), a
    percentage progress, a progress bar and an adaptative ETA.

    Args:
        iterator (iterator): Iterator of items to use the bar with.
        text (str): Text to display describing the current operation.

    Returns:
        generator object: Item handled by the progress bar.
    """
    widgets = []

    # add optional text widget
    if text:
        widgets.extend([ShrinkableTextWidget(text), " "])

    # add other widgets
    widgets.extend(
        [
            progressbar.Percentage(),
            " ",
            progressbar.Bar(),
            " ",
            progressbar.Timer(),
            " ",
            progressbar.AdaptiveETA(),
        ]
    )

    # create progress bar
    with progressbar.ProgressBar(*args, widgets=widgets, **kwargs) as progress:
        for item in progress(iterator):
            yield item


def null_bar(iterator, *args, text=None, **kwargs):
    """Generator that gives the defaylt muted progress bar for the project.

    It only logs the optionnal text.

    Args:
        iterator (iterator): Iterator of items to use the bar with.
        text (str): Text to log describing the current operation.

    Returns:
        generator object: Item handled by the progress bar.
    """
    # log text immediately
    if text:
        logger.info(text)

    # create null progress bar
    with progressbar.NullBar(*args, **kwargs) as progress:
        for item in progress(iterator):
            yield item



"""Config helper module.

This module gives the class `Config` that handles the config. It can load
values from a YAML file and from environment variables:

>>> from path import Path
>>> config = Config("DAKARA")
>>> config.load_file(Path("path/to/file.yaml"))
>>> config.set_debug()
>>> config.check_mandatory_key("server")
>>> config.get("server").get("address")

The module has two functions to configure loggers: `create_logger`, which
installs the logger using coloredlogs, and `set_loglevel`, which sets the
loglevel of the logger according to the config. Usually, you call the first one
before reading the config, as `load_config` needs a logger, then call the
latter one:

>>> create_logger()
>>> from path import Path
>>> config = Config("DAKARA")
>>> config.load_file(Path("path/to/file.yaml"))
>>> set_loglevel(config)

If you use progress bar and logging at the same time, you should call
`create_logger` with `wrap=True`.

The module has one function to manage Dakara Project config files,
`create_config_file`, that copies a given config file stored in module resources to
the configuration directory:

>>> create_config_file("module.resources", "my_config.yaml")
"""


import logging
from collections import UserDict
from distutils.util import strtobool

import coloredlogs
import progressbar
import yaml
from environs import Env, EnvError
from path import Path

try:
    from importlib.resources import path

except ImportError:
    from importlib_resources import path

from dakara_base.directory import directories
from dakara_base.exceptions import DakaraError

LOG_FORMAT = "[%(asctime)s] %(name)s %(levelname)s %(message)s"
LOG_LEVEL = "INFO"

logger = logging.getLogger(__name__)


class AutoEnv(Env):
    """Environment variable reader with an automatic method."""

    def auto(self, type, *args, **kwargs):
        type_str = type.__name__
        return getattr(self, type_str)(*args, **kwargs)


class Config(UserDict):
    """Configuration object.

    This object behaves similarly to a dictionary. Its values can be populated
    from a file at first. Then, when accessing a value with a key, it first
    checks if a value with similar name exists as an environment variable. In
    that case, the environment variable is given, otherwise the stored value is
    given.

    Values can be loaded from a file, previous values stored in the config
    would be discarded.

    >>> from path import Path
    >>> conf = Config("prefix")
    >>> conf.load_file(Path("config.yaml"))

    When checking environment variables, the looked up variable name is
    prefixed and made upper-case.

    >>> conf = Config("prefix", {"key1": "foo", "key2": "bar"})
    >>> conf
    {"key1": "foo", "key2": "bar"}
    >>> conf.get("key1")
    "foo"
    >>> # let's say PREFIX_KEY2 is an environment variable with value "spam"
    >>> conf.get("key2")
    "spam"

    Values of nested `Config` objects will have accumulated prefixes
    (separated by an underscore):

    >>> conf = Config("prefix", {"sub": {"key": "foo"}})
    >>> # let's say PREFIX_SUB_KEY is an environment variable with value "bar"
    >>> cong.get("sub").get("key")
    "bar"

    By default, the value obtained from the environment is a string. If a
    default value is provided to `get`, the returned value from the environment
    will be parsed to the type of that default value:

    >>> conf = Config("prefix", {"key": 42})
    >>> # let's say PREFIX_KEY is an environment variable with value "39"
    >>> conf.get("key")
    "39"
    >>> cong.get("key", 0)
    39

    You can check `environs.Env` for the supported types. Note stored values
    are parsed from the config file by the YAML library.

    Attributes:
        prefix (str): Prefix to use when looking for value in environment
            variables.
        env (AutoEnv): Environment parser.

    Args:
        prefix (str): Prefix to use when looking for value in environment
            variables.
        iterable (iterable): Values to store.
    """

    def __init__(self, prefix, iterable=None):
        super().__init__()

        self.prefix = prefix
        self.env = AutoEnv()

        # create values in object if any provided
        if iterable:
            self.set_iterable(iterable)

    def set_iterable(self, iterable):
        """Set config values from the provided iterable.

        Dictionaries will be converted into Config with a sub-prefix.

        Args:
            iterable (dict): Dictionary of values.
        """
        # recursively convert dictionaries into Config objects
        iterable = {
            key: (
                self.__class__("{}_{}".format(self.prefix, key), val)
                if isinstance(val, dict)
                else val
            )
            for key, val in iterable.items()
        }

        # reset config data with values from the iterable
        self.data.clear()
        self.data.update(iterable)

    def set_debug(self, debug=True):
        """Set log level of the config to debug.

        Args:
            debug (bool): If `True` (default), set log level to "DEBUG".
        """
        if debug:
            self.data["loglevel"] = "DEBUG"

    def check_mandatory_keys(self, keys):
        """Check if a list of keys is present in the config.

        Args:
            keys (list of str): Keys that must be present in the config.
        """
        for key in keys:
            self.check_mandatory_key(key)

    def check_mandatory_key(self, key):
        """Check if a key is present in the config.

        Args:
            keys (str): Key that must be present in the config.

        Raises:
            ConfigInvalidError: If the config misses a critical section.
        """
        if key not in self.data:
            raise ConfigInvalidError("Invalid config file, missing '{}'".format(key))

    def load_file(self, config_path):
        """Load config from a given YAML file.

        Args:
            config_path (path.Path): Path to the config file.

        Raises:
            ConfigNotFoundError: If the config file cannot be open.
            ConfigParseError: If the config cannot be parsed.
        """
        logger.info("Loading config file '%s'", config_path)

        # load and parse the file and create config data
        try:
            with config_path.open() as file:
                self.set_iterable(yaml.safe_load(file))

        except yaml.parser.ParserError as error:
            raise ConfigParseError("Unable to parse config file") from error

        except FileNotFoundError as error:
            raise ConfigNotFoundError("No config file found") from error

    def get_value_from_env(self, key, type=None):
        """Get the value from prefixed upper case environment variable.

        Args:
            key (str): Name of the variable without prefix.
            type (type): Type of the variable. If not provided, default to
                string.

        Returns:
            str: Value from environment variable.
        """
        with self.env.prefixed("{}_".format(self.prefix.upper())):
            # use type if provided
            if type:
                return self.env.auto(type, key.upper())

            # fallback to default behavior
            return self.env(key.upper())

    def __getitem__(self, key):
        # try to get value from environment
        try:
            return self.get_value_from_env(key)

        except EnvError:
            return super().__getitem__(key)

    def get(self, key, default=None):
        """Return the value for the provided key.

        If a default value is provided, it will determine the type of the
        returned value when getting it from the environment variables.

        Args:
            key (any): Key to retreive.
            default (any): Default value if the key cannot be found.

        Returns:
            any: Value. If `default` was provided, it will be of the same type.
        """
        # guess cast from default value
        cast = None
        if default is not None:
            cast = type(default)

        # get value from environment, then from dict
        try:
            return self.get_value_from_env(key, cast)

        except EnvError:
            return super().get(key, default)


def create_logger(wrap=False, custom_log_format=None, custom_log_level=None):
    """Create logger.

    Args:
        wrap (bool): If True, wrap the standard error stream for using logging
            and progress bar. You have to enable this flag if you use
            `progress_bar`.
        custom_log_format (str): Custom format string to use for logs.
        custom_log_level (str): Custom level of logging.
    """
    # wrap stderr on demand
    if wrap:
        progressbar.streams.wrap_stderr()

    # setup loggers
    log_format = custom_log_format or LOG_FORMAT
    log_level = custom_log_level or LOG_LEVEL
    coloredlogs.install(fmt=log_format, level=log_level)


def set_loglevel(config):
    """Set logger level.

    Arguments:
        config (Config): Dictionary of the config.
    """
    loglevel = config.get("loglevel", LOG_LEVEL)
    coloredlogs.set_level(loglevel)


def create_config_file(resource, filename, force=False):
    """Create a new config file in user directory.

    Args:
        resource (str): Resource where to find the config file.
        filename (str): Name of the config file.
        force (bool): If True, config file in user directory is overwritten if
            it existed already. Otherwise, prompt the user.
    """
    with path(resource, filename) as file:
        # get the file
        origin = Path(file)
        destination = directories.user_config_dir / filename

        # create directory
        destination.dirname().mkdir_p()

        # check destination does not exist
        if not force and destination.exists():
            try:
                result = strtobool(
                    input("{} already exists, overwrite? [y/N] ".format(destination))
                )

            except ValueError:
                result = False

            if not result:
                return

        # copy file
        origin.copyfile(destination)
        logger.info("Config created in '{}'".format(destination))


class ConfigError(DakaraError):
    """Generic error raised for invalid configuration file."""


class ConfigNotFoundError(ConfigError):
    """Unable to read configuration file."""


class ConfigParseError(ConfigError):
    """Unable to parse config file."""


class ConfigInvalidError(ConfigError):
    """Config has missing mandatory keys."""



"""Exceptions helper module.

This module defines the base exception class for any project using this
library. All exception classes should inherit from `DakaraError`:

>>> class MyError(DakaraError):
...     pass

This helps to differentiate known exceptions and unknown ones, which are real
bugs.

It defines `generate_exception_handler` that allows to generate functions which
will catch an exception, add a custom error message and re-raise it. It also
defines `handle_all_exceptions` that can be used on `__main__` module to catch
all program exceptions.
"""

import logging
from contextlib import contextmanager

logger = logging.getLogger(__name__)


class DakaraError(Exception):
    """Basic exception class for the project."""


class DakaraHandledError(Exception):
    """Basic exception class for errors that have been handled.

    Must be used in multiple inheritance.
    """


def generate_exception_handler(exception_class, error_message):
    """Generate a context manager to take care of given exception.

    It will add a custom message to an expected exception class. An exception
    derived from the caught exception (that can be a subclass of
    `exception_class`) and from the generic `DakaraHandledError` is then
    raised.

    >>> class MyError(Exception):
    ...     pass
    >>> handle_my_error = generate_exception_handler(MyError, "extra message")
    >>> try:
    ...     with handle_my_error():
    ...         raise MyError("initial message")
    ... except MyError as error:
    ...     pass
    >>> assert str(error).split() == ["initial message", "extra message"]
    >>> assert isinstance(error, MyError)
    >>> assert isinstance(error, DakaraHandledError)

    Args:
        exception_class (Exception or list of Exception): Exception class to
            catch.
        error_message (str): Error message to display. It will be displayed on
            the next line after the exception message.

    Returns:
        function: Context manager function.
    """

    @contextmanager
    def function():
        try:
            yield None

        except exception_class as error:

            class HandledError(error.__class__, DakaraHandledError):
                pass

            raise HandledError("{}\n{}".format(error, error_message)) from error

    return function


class ExitValue:
    """Container for the exit value.

    Attributes:
        value (int): Exit value, default to 0.
    """

    def __init__(self):
        self.value = 0


@contextmanager
def handle_all_exceptions(bugtracker_url, logger=logger, debug=False):
    """Handle all exceptions and yield an exit value.

    Unless in debug mode, no exceptions will be raised.

    >>> import sys
    >>> with handle_all_exceptions(
    ...    "https://www.example.com/bugtracker"
    ... ) as exit_value:
    ...    # your program here
    >>> sys.exit(exit_value.value)

    Args:
        bugtracker_url (str): URL address of the bugtracker, displayed on
            unexpected exceptions.
        logger (logging.Logger): Logger. If not given, will take the current
            module's logger.
        debug (bool): If True, known and unknown exceptions will be directly
            raised.

    Yields:
        ExitValue: Container with the return value, stored in attribute
        `value`. If no error happened, the return value is 0, in case of
        Ctrl+C, it is 255, in case of a known error, it is 1, in case of an
        unknown error, it is 2.
    """
    container = ExitValue()

    try:
        yield container

    except KeyboardInterrupt:
        container.value = 255
        logger.info("Quit by user")

    except DakaraError as error:
        container.value = 1

        # directly re-raise the error in debug mode
        if debug:
            raise

        # just log it otherwise
        logger.critical(error)

    except BaseException as error:
        container.value = 2

        # directly re-raise the error in debug mode
        if debug:
            raise

        # show the error and a special message otherwise
        logger.exception("Unexpected error: %s", error)
        logger.critical("Please fill a bug report at '%s'", bugtracker_url)
